"""
Classes and methods to manage tag attributes
"""
import regex

from .. import settings
from ..utils.list import uniq
from ..utils.types import PositionalValue, positionalvalue_type, ispositional


class AttributeFormatError(Exception):
    """An error was encountered in the format of attributes"""
    pass


class _MissingAttribute(object):
    """An internal object used when a key is missing"""
    pass


def strip_attr(attr, sep=settings.attribute_target_sep):
    """Strip a target-specific terminator from the given attr.

    Parameters
    ----------
    attr : Union[str, :class:`PositionalValue <.PositionalValue>`]
            The key of the attribute to retrieve or the PositionalValue to
            retrieve.
    sep : Optional[str]
        The separator character (or string) to use to separate the key
        and target.

    Returns
    -------
    stripped_attr : str
        The attribute with the target-specific terminator stripped.

    Examples
    --------
    >>> strip_attr(3)
    3
    >>> strip_attr('143')
    '143'
    >>> strip_attr('width.tex')
    'width'
    >>> strip_attr('width')
    'width'
    >>> strip_attr('my_file.pdf#link-anchor')
    'my_file.pdf#link-anchor'
    >>> strip_attr('my_file.pdf#link-anchor.tex')
    'my_file.pdf#link-anchor'
    """
    if isinstance(attr, str):
        # If a target-specific terminator is not included, nothing needs to
        # be done
        if sep not in attr:
            return attr

        # Check all available targets
        targets = {target.strip('.')
                   for target in settings.tracked_deps.keys()}
        for target in targets:
            term = sep + target
            if attr.endswith(term):
                return attr[:-len(term)]

    return attr


def is_target_specific(attr, sep=settings.attribute_target_sep):
    """Tests whether the given attribute is a target-specific attribute.

    Parameters
    ----------
    attr : Union[str, :class:`PositionalValue <.PositionalValue>`]
            The key of the attribute to retrieve or the PositionalValue to
            retrieve.
    sep : Optional[str]
        The separator character (or string) to use to separate the key
        and target.

    Returns
    -------
    is_target_specific : bool
        True if the attr is a target-specific attribute

    Examples
    --------
    >>> is_target_specific(3)
    False
    >>> is_target_specific('pos')
    False
    >>> is_target_specific('pos.tex')
    True
    >>> is_target_specific('/usr/docs/text.tex')  # not quoted
    True
    >>> is_target_specific('"/usr/docs/text.tex"')  # quoted
    False
    """
    return attr != strip_attr(attr, sep=sep)


re_attrs = regex.compile(r'((?P<key>[\w\.]+)'
                         r'\s*=\s*'
                         r'(?P<value>("[^"]*"'
                         r'|\'[^\']*\''
                         r'|[^\s\]]+))'
                         r'|(?P<position>"[^"]*"'
                         r'|\'[^\']*\''
                         r'|[^\s,\[\]]+|))')


class Attributes(dict):
    """The attributes class is an ordered dict to manage and validate attribute
    entries.

    .. note:: Attributes can either be key/value (ex: 'class=media') or
              positional (ex: 'red'). For positional arguments, the values
              are :class:`PositionalValue <.utils.types.PositionalValue>` class
              objects.
    """

    def __init__(self, *args, **kwargs):
        for string in filter(lambda x: isinstance(x, str), args):
            self.load(string)

        # Remove strings (and None) from args
        args = tuple(filter(lambda x: not isinstance(x, str) and x is not None,
                            args))

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return self.__class__.__name__ + super().__repr__()

    def copy(self):
        return Attributes(**self)

    def totuple(self):
        """Return a tuple for an attributes dict

        Examples
        --------
        >>> attrs = Attributes("key1=val1 val2")
        >>> attrs.totuple()
        (('key1', 'val1'), 'val2')
        """
        return tuple(k if ispositional(v) else (k, v)
                     for k, v in self.items())

    def load(self, s, sep=settings.attribute_target_sep):
        """Parses an attribute string into key/values for the Attributes dict.

        Parameters
        ----------
        s: str
            Input string of attributes. Attributes come in two forms:
            positional attributes and keyword attributes (kwargs), and
            attributes strings have the following form:
            "key1=value1 key2=value2 value3"
        sep : Optional[str]
            The separator character (or string) to use to separate the key
            and target.

        Examples
        --------
        >>> attrs = Attributes()
        >>> attrs.load("key1=val1 val2")
        >>> attrs
        Attributes{'key1': 'val1', 'val2': <class '...StringPositionalValue'>}
        """
        attrs = []

        for m in re_attrs.finditer(s):
            d = m.groupdict()
            if d.get('key', None) and d.get('value', None):
                # Put the key-value pair in the attributes dict
                attrs.append((d['key'], d['value'].strip('"').strip("'")))

            elif d.get('position', None):
                value = d['position'].strip("'").strip('"')

                # Strip target-specific suffixes to get the accurate
                # PostionalValue subclass. (ex: '3.1416.tex' turns into
                # '3.1416' to create a FloatPositionalValue instead of just a
                # PositionalValue)
                gen_value = strip_attr(value, sep=sep)

                # Add the positional attribute to the dict by putting the
                # positional value as a key and the PostionalValue class or
                # subclass as the value
                attrs.append((value, positionalvalue_type(gen_value)))

        self.update(tuple(attrs))

    def find_item(self, attr, target=None, default=None,
                  sep=settings.attribute_target_sep):
        """Find the key in the attributes dict with or without the target
        specified.

        This function will not modify the keys/values in this attributes dict.

        Parameters
        ----------
        attr : Union[str, :class:`PositionalValue <.PositionalValue>`]
            The key of the attribute to retrieve or the PositionalValue to
            retrieve.
        target : Optional[str]
            If specified, search for a key for the given target.
            ex: class.html with be returned for the 'class' key and 'html'
            target, if available, otherwise, the entry for 'class' will be
            returned.
        default : Optional[Any]
            If the key is not found, return this value instead.
        sep : Optional[str]
            The separator character (or string) to use to separate the key
            and target.

        Returns
        -------
        key, value : Any
            The retrieved item.

        Examples
        --------
        >>> attrs = Attributes('class="general" class.html="specific" pos')
        >>> attrs.find_item('class')
        ('class', 'general')
        >>> attrs.find_item('class.html')
        ('class.html', 'specific')
        >>> attrs.find_item('class', target='html')
        ('class.html', 'specific')
        >>> attrs.find_item('class', target='tex')
        ('class', 'general')
        >>> attrs.find_item('pos')
        ('pos', <class ...StringPositionalValue'>)
        >>> attrs.find_item('missing')
        """
        target = target.strip('.') if isinstance(target, str) else target

        # See if the target specific key is in the attributes
        # ex: 'class.html' for target='.html' or 'file.txt.tex'
        target_key = (sep.join((str(attr), target)) if target is not None
                      else None)
        if target_key and target_key in self:
            return target_key, self[target_key]

        # See if the attr is a key in the dict itself. ex: 'class'
        if attr in self:
            return attr, self[attr]

        # See if it's a positional value
        if ispositional(attr):
            # Find all matching positional values
            # ex: StringPositionalValue
            items = [(k, v) for k, v in self.items() if ispositional(v, attr)]

            if target:
                # Filter by target, if specific
                target_items = [(k, v) for k, v in items if k.endswith(target)]
            else:
                # Otherwise remove entries that have target-specific
                target_items = [(k, v) for k, v in items
                                if not is_target_specific(k, sep=sep)]

            # Return the first item found
            if target_items:
                return target_items[0]
            elif items:
                return items[0]

        # Give up
        return default

    def get(self, attr, target=None, default=None,
            sep=settings.attribute_target_sep):
        """Retrieve an entry by target-specific key, if available and
        specified, or by key.

        As opposed to the find_item method, this function will strip the
        target-specific terminators of returned values

        Parameters
        ----------
        attr : Union[str, :class:`PositionalValue <.PositionalValue>`]
            The key of the attribute to retrieve or the PositionalValue to
            retrieve.
        target : Optional[str]
            If specified, search for a key for the given target.
            ex: class.html with be returned for the 'class' key and 'html'
            target, if available, otherwise, the entry for 'class' will be
            returned.
        default : Optional[Any]
            If the key is not found, return this value instead.
        sep : Optional[str]
            The separator character (or string) to use to separate the key
            and target.

        Returns
        -------
        value : Any
            The retrieved value.
            If the entry is a positional attribute, PositionalValue will be
            returned.

        Examples
        --------
        >>> attrs = Attributes('class="general" class.html="specific" pos')
        >>> attrs.get('class')
        'general'
        >>> attrs.get('class.html')
        'specific'
        >>> attrs.get('class', target='html')
        'specific'
        >>> attrs.get('class', target='tex')
        'general'
        >>> attrs.get('pos')
        <class ...StringPositionalValue'>
        >>> attrs.get('missing')
        """
        rv = self.find_item(attr=attr, target=target, default=default,
                            sep=sep)
        if isinstance(rv, tuple) and len(rv) == 2:
            key, value = rv
            rv = key if ispositional(attr) else value

            return strip_attr(rv, sep=sep) if isinstance(rv, str) else rv
        else:
            return rv

    def append(self, attr, value=_MissingAttribute):
        """Set a value by appending to it, if it already exists, or creating
        it if it doesn't.

        Parameters
        ----------
        attr : str
            The attr (key) for the entry to set or append.
        value : Optional[Any]
            The value to set or append. If not specified, the attribute will
            be added as a positional argument
        """
        if value == _MissingAttribute:
            # Simply add positional attributes
            self[attr] = PositionalValue
        elif attr not in self:
            # Simply set the value, if it doesn't exist
            self[attr] = value
        else:
            current_value = self[attr]
            # In this case, try to append it
            if isinstance(current_value, str):
                self[attr] = ' '.join((current_value, str(value)))
            elif isinstance(current_value, list):
                current_value.append(value)
            else:
                # Append it as a string
                self[attr] = ' '.join((str(current_value), str(value)))

    def strip(self, sep=settings.attribute_target_sep):
        """Replace the entries in this attributes dict without the
        target-specific terminators."""
        # Make a copy and reset this dict
        cp = self.copy()
        self.clear()

        # Get all the keys to copy over
        keys = list(cp.keys())

        # Find target-specific terminators from keys.
        stripped_keys = set()
        for key in keys:
            stripped_key = strip_attr(key, sep=sep)
            if stripped_key != key:
                stripped_keys.add(stripped_key)

        # Remove keys that have target-specific entries
        keys = [k for k in keys if k not in stripped_keys]

        for key in keys:
            stripped_key = strip_attr(key, sep=sep)

            if key == stripped_key:
                # use the key, if it hasn't been changed
                self[key] = cp[key]
            else:
                self[stripped_key] = cp[key]

    def filter(self, attrs=None, target=None,
               sep=settings.attribute_target_sep, sort_by_attrs=False,
               strip=True):
        """Create an Attributes dict with target-specific entries.

        Parameters
        ----------
        attrs : Optional[Union[str, List[Union[str, :class:`PositionalValue \
            <.PositionalValue>`]]]]
            Filter keys. If specified, only entries that match one of these
            keys will be returned.
        target : Optional[str]
            Filter targets. If specified, only entries general entries will
            be returned unless a target-specific entry is present, in which
            case it will be returned.
            For example, if entries for 'class' and 'class.tex' exist and
            target='tex', then the 'class.tex' entry will be returned as
            'class'. If target is None, then the 'class' entry will be
            returned.
        sep : Optional[str]
            The separator character (or string) to use to separate the key
            and target.
        sort_by_attrs : Optional[bool]
            If True, the returned attributes dict will have keys sorted in the
            same order as the attrs passed. Otherwise, the returned attributes
            dict will be sorted with keys in the same order as this attributes
            dict (default).
        strip : Optional[bool]
            If True, strip all target-specific terminators from the returned
            attris dict.

        Returns
        -------
        attributes : :obj:`Attributes <.attributes.Attributes>`
            A filtered attributes dict.

        Examples
        --------
        >>> attrs = Attributes('class=one tgt=default tgt.tex=tex')
        >>> attrs.filter()  # python >= 3.6 should have ordered entries
        Attributes{'class': 'one', 'tgt': 'default'}
        >>> attrs.filter(target='tex', strip=False)
        Attributes{'class': 'one', 'tgt.tex': 'tex'}
        >>> attrs.filter(target='tex', strip=True)
        Attributes{'class': 'one', 'tgt': 'tex'}
        >>> attrs.filter(attrs='class')
        Attributes{'class': 'one'}
        >>> attrs.filter(attrs='tgt')
        Attributes{'tgt': 'default'}
        """
        target = target.strip('.') if isinstance(target, str) else target

        # Setup the returned attributes dict
        d = Attributes()

        # Setup the attrs without target-specific values
        attrs = ([attrs] if isinstance(attrs, str) or
                 isinstance(attrs, PositionalValue)
                 else attrs)  # wrap strings
        attrs = (list(self.keys())
                 if attrs is None else attrs)  # populate empty attrs
        attrs = uniq(attrs)  # general attrs

        if not sort_by_attrs:
            # Sort the attrs so that they follow the same order as keys in
            # this dict. This ensures the results of filter are deterministic
            # and not random, if the attrs are an unordered iterable, like a
            # set
            order = {k: num for num, k in enumerate(self.keys())}
            attrs = sorted(attrs, key=lambda x: order.get(x, len(order)))

        # Find keys to remove
        if target is not None:
            # Find target-specific terminators from keys.
            remove_attrs = set()
            for attr in attrs:
                stripped_attr = strip_attr(attr, sep=sep)

                if attr == stripped_attr:
                    # If the attr does not have a target-specific terminator,
                    # do not remove it.
                    continue
                # The following are entries when attr != stripped_attr, i.e.
                # key had a target-specific terminator
                if attr.endswith(sep + target):
                    # The attr is a target-specific attr that matches the
                    # target. Remove the non-target-specific general entry
                    remove_attrs.add(stripped_attr)
                else:
                    # The attr is a target-specific attr, but it matches a
                    # different target. Remove that.
                    remove_attrs.add(attr)
        else:
            # Remove target-specific keys
            remove_attrs = {attr for attr in attrs
                            if is_target_specific(attr, sep=sep)}

        # Remove those entries
        attrs = [attr for attr in attrs if attr not in remove_attrs]

        for attr in attrs:
            rv = self.find_item(attr=attr, target=target, sep=sep)
            if rv is not None:
                k, v = rv
                d[k] = v

        # Strip target-specific terminators, if strip is enabled
        if strip:
            d.strip()
        return d

    @property
    def html(self):
        """Format the attributes for html.

        Notes
        -----
        - This function won't strip target tags ('.html'), and a filter
          for the '.html' target should be applied before using this property

        """
        # Disable filtering here because the command that uses this property
        # already filters the attributes dict
        # d = self.filter(target='html')  # previously
        d = self

        # Create an attribute string in html format
        entries = ["{}='{}'".format(k, v) if not ispositional(v) else
                   "{}".format(k)
                   for k, v in d.items()]
        return " ".join(entries)

    @property
    def tex_arguments(self):
        """Format arguments for tex.

        Notes
        -----
        - This function won't strip target tags ('.tex'), and a filter
          for the '.tex' target should be applied before using this property

        Examples
        --------
        >>> Attributes('width=302 3').tex_arguments
        '{302}{3}'
        """
        # Disable filtering here because the format command that uses this
        # property already filters the attributes dict
        # d = self.filter(target='tex')  # previously
        d = self

        # Create an attribute in tex format
        entries = ["{{{}}}".format(v)
                   if not ispositional(v) else
                   "{{{}}}".format(k)
                   for k, v in d.items()]
        return "".join(entries) if entries else ""

    @property
    def tex_optionals(self):
        """Format optional arguments for tex.

        Notes
        -----
        - This function won't strip target tags ('.tex'), and a filter
          for the '.tex' target should be applied before using this property

        Examples
        --------
        >>> Attributes('width=302 3').tex_optionals
        '[width=302, 3]'
        """
        # Disable filtering here because the command that uses this property
        # already filters the attributes dict
        # d = self.filter(target='tex')  # previously
        d = self

        # Create an attribute in tex format
        entries = ["{}={}".format(k, v)
                   if not ispositional(v) else
                   "{}".format(k)
                   for k, v in d.items()]

        # Format the optional arguments. Treat '*' outside of parentheses
        return "[" + ", ".join(entries) + "]" if entries else ""



"""
A function to register content labels
"""
from .types import ContentLabel, DocumentLabel


def register_content_labels(labels, **kwargs):
    """Assign chapter/section/subsection links for Content labels.

    This processor only works on content labels (:obj:`ContentLabel
    <.types.ContentLabel>`)

    Parameters
    ----------
    labels : Dict[Tuple[str,str], :obj:`Label <.label_manager.types.Label>`]
        A list of labels where the key is the (doc_id, label_id) and the values
        are the label objects.
    """

    # Keep track of the current heading registered_labels
    part_label = None
    chapter_label = None
    section_label = None
    subsection_label = None
    subsubsection_label = None

    # Process each Contentlabel.
    content_labels = [label for label in labels.values()
                      if isinstance(label, ContentLabel)]

    # Keep track of the current doc_id
    doc_id = None

    for label in content_labels:
        # Switch the local count whenever a new document is encountered.
        # However, the local_counter is not reset between documents
        if not isinstance(label, DocumentLabel) and doc_id != label.doc_id:
            doc_id = label.doc_id

        if label.kind and label.kind[-1] == 'part':
            # If it's a new chapter (i.e. a chapter or document title)
            # Get its label and reset the counters for the heading
            # registered_labels below
            part_label = label
            chapter_label = None
            section_label = None
            subsection_label = None
            subsubsection_label = None

        elif label.kind and label.kind[-1] == 'chapter':
            # If it's a new chapter (i.e. a chapter or document title)
            # Get its label and reset the counters for the heading
            # registered_labels below
            chapter_label = label
            section_label = None
            subsection_label = None
            subsubsection_label = None

        elif label.kind and label.kind[-1] == 'section':
            section_label = label
            subsection_label = None
            subsubsection_label = None

        elif label.kind and label.kind[-1] == 'subsection':
            subsection_label = label
            subsubsection_label = None

        elif label.kind and label.kind[-1] == 'subsubsection':
            subsubsection_label = label

        # Set the heading label for all registered_labels
        label.part_label = part_label
        label.chapter_label = chapter_label
        label.section_label = section_label
        label.subsection_label = subsection_label
        label.subsubsection_label = subsubsection_label



"""
The Checker base class for checking dependencies.
"""
import shutil

from .types import All, SoftwareDependency, SoftwareDependencyList
from ..utils.classes import all_subclasses
from .exceptions import MissingHandler


class Checker(All):
    """Check for installed dependencies.

    .. note::
        The check implements checker handlers for different categories of
        dependencies (``check_`` methods).

        For example, a dependencies listing with a category 'executables' will
        run the :meth:`check_executables` method and a listing with a category
        of 'packages' will run the :meth:`check_packages` method.
    """

    @classmethod
    def checker_subclasses(cls):
        """All subclasses checkers."""
        return all_subclasses(cls)

    @classmethod
    def find_executable(cls, executable):
        return shutil.which(executable)

    def check(self):
        # Find SoftwareDependencyList categories.
        categories = [i.category for i in self.dependencies
                      if isinstance(i, SoftwareDependencyList)]
        for category in categories:
            handler_name = 'check_' + category
            meth = getattr(self, handler_name, None)
            if meth is not None:
                meth()
            else:
                msg = "The checker '{}' is missing the handler '{}'."
                cls_name = self.__class__.__name__

                raise MissingHandler(msg.format(cls_name, handler_name))

        return self.available

    def check_executables(self, executables=None):
        """Check the availability of executables.

        Parameters
        ----------
        executables : Union[List[str], Tuple[str]]
            The list of executables to check for availability.
        """
        # Retrieve the listing of executables
        if executables is not None:
            pass
        elif 'executables' in self:
            executables = self['executables']
        else:
            return None

        # Iterate over the executables and see if they can be found.
        for e in executables:
            if isinstance(e, SoftwareDependency):
                # Skip if already processed, as 'which' (below) might be
                # expensive to process
                if e.available is not None:
                    continue

                # If it's a software dependency, see if it's available
                path = shutil.which(e.name)
                e.available = path is not None  # executable found
                e.path = path
            elif isinstance(e, SoftwareDependencyList):
                # If it's a software dependency list, check each of those items
                self.check_executables(e.dependencies)

    def check_packages(self, packages=None):
        """Check the availability of packages.

        Parameters
        ----------
        packages : Union[List[str], Tuple[str]]
            The list of packages to check for availability.
        """
        # Get a listing of 'check_packages' methods
        check_packages_meth_names = [meth for meth in dir(self)
                                     if meth.startswith('check_packages_')]

        # Execute the methods
        for meth_name in check_packages_meth_names:
            meth = getattr(self, meth_name)
            meth(packages)



"""
A TargetBuilder for txt files.
"""
from .target_builder import TargetBuilder


class TxtBuilder(TargetBuilder):
    """A builder for txt files."""

    available = True
    priority = 1000
    infilepath_ext = '.dm'
    outfilepath_ext = '.txt'



"""
Tags for document preambles.
"""
from .tag import Tag
from .headings import Title
from ..formats import xhtml_tag
from ..utils.string import str_to_list


class Authors(Tag):
    """A tag for listing the author or authors."""

    aliases = ('author',)
    active = True

    def __init__(self, name, context, **kwargs):
        super(Authors, self).__init__(name=name, context=context, **kwargs)

        # Use the specified author list in the content, if specified, otherwise
        # try to get it from the context.
        if (isinstance(self.content, str) and self.content.strip() == "" or
           self.content is None):

            author = (context.get('author', '') if 'author' in context else
                      context.get('authors', ''))
            self.content = author

        # Convert the author string to a list
        if isinstance(self.content, str):
            self.content = str_to_list(self.content)

    def author_string(self):
        """Generate a formatted string listing the authors."""
        # Convert to a list of strings
        if isinstance(self.content, str):
            author_lst = str_to_list(self.content)
        elif isinstance(self.content, list):
            author_lst = self.content
        else:
            return ''

        # Convert to a list of authors
        if len(author_lst) == 0:
            return ''
        elif len(author_lst) == 1:
            return author_lst[0]
        else:
            last_two = author_lst[-2:]
            others = author_lst[:-2]

            if len(author_lst) == 2:
                return ', '.join(others) + ' and '.join(last_two)
            else:
                return ', '.join(others) + ', ' + ' and '.join(last_two)

    def tex_fmt(self, *args, **kwargs):
        return self.author_string()

    def html_fmt(self, content=None, method='html', level=1, *args, **kwargs):
        return xhtml_tag('span', attributes='class=authors',
                         formatted_content=self.author_string(),
                         method=method, level=level)


class Titlepage(Tag):
    """A titlepage tag."""

    authors_tag = None
    active = True

    def __init__(self, name, content, attributes, context):
        super(Titlepage, self).__init__(name, content, attributes, context)

        # Setup the title tag
        self.title_tag = Title(name='title', content='', attributes='',
                               context=context)

        # Setup the author tag
        self.authors_tag = Authors(name='authors', content='',
                                   attributes=tuple(), context=context)

    @property
    def title(self):
        """The title of the project"""
        return self.context.get('title', '')

    @property
    def authors(self):
        """The authors of the document"""
        if 'document' in self.context:
            doc = self.context['document']
            if hasattr(doc, 'author'):
                return doc.author
        if 'author' in self.context:
            return self.context['author']

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1, **kwargs):
        title_tag = self.title_tag
        title_html = getattr(title_tag, format_func)(method=method,
                                                     level=level + 1)
        authors_tag = self.authors_tag
        authors_html = getattr(authors_tag, format_func)(content=content,
                                                         method=method,
                                                         level=level + 1)
        return xhtml_tag('div', attributes='class=title-page',
                         formatted_content=[title_html, authors_html],
                         method=method, level=level)

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        return "\n\\maketitle\n\n"



"""
String manipulation operations.
"""
import hashlib
from itertools import groupby

import regex
from slugify import slugify  # noqa: F401

from .. import settings


def hashtxt(text, truncate=10):
    """Creates a hash from the given text."""
    text = text if isinstance(text, bytes) else text.encode()
    return (hashlib.md5(text).hexdigest() if truncate is None else
            hashlib.md5(text).hexdigest()[:truncate])


def titlelize(string, truncate=True, capitalize=False):
    """Given a string, generate a condensed title.

    >>> titlelize("My example caption. It has 2 sentences")
    'My example caption'
    >>> titlelize("My example caption. It has 2 sentences", capitalize=True)
    'My Example Caption'
    """
    # Strip extra newlines and join to a single string
    lines = list(filter(bool, string.split('\n')))
    string = ' '.join(lines)

    # Break at end of sentence periods
    pieces = string.split('. ')

    # Truncate, if needed
    if truncate:
        string = pieces[0]
    else:
        string = ". ".join(pieces)

    # capitalize words, if specified, and return the joined string.
    # This function also strips extra spaces between words
    pieces = string.split(' ')
    pieces = list(filter(bool, pieces))
    if capitalize:
        if len(pieces) > 2:
            pieces = ([pieces[0].title()] +
                      [piece.title() if piece not in ('in', 'to', 'a')
                       else piece for piece in pieces[1:]])
    string = " ".join(pieces)

    return string.strip('.')


def nicejoin(*items, sep=', ', term=' and '):
    """Join a sequence of strings with a separator and an alternative terminal
    separator.

    Parameters
    ----------
    *items : Tuple[str]
        A sequence of strings to join.
    sep : Optional[str]
        The separator to use for all items, other than the last item.
    term : Optional[str]
        The last separator to use in joining items.

    Returns
    -------
    joined_string : str
        The joined string.

    Examples
    --------
    >>> nicejoin('Bill', 'Ted', 'Frances')
    'Bill, Ted and Frances'
    >>> nicejoin('Bill', 'Ted', 'Frances', term=' or ')
    'Bill, Ted or Frances'
    >>> nicejoin('Bill', 'Ted', 'Frances')
    'Bill, Ted and Frances'
    >>> nicejoin('Bill', 'Ted')
    'Bill and Ted'
    >>> nicejoin('Bill')
    'Bill'
    """
    if len(items) > 1:
        return sep.join(items[:-1]) + term + items[-1]
    elif len(items) == 1:
        return items[0]
    else:
        return ''


def space_indent(s, number=4):
    r"""Indent a text block by the specified number of spaces.

    Parameters
    ----------
    s : str
        The string the indent
    number : Optional[int]
        The number of spaces to indent by.

    Returns
    -------
    indented_string : str
        The indented string.

    Examples
    --------
    >>> space_indent('my test')
    '    my test'
    >>> t=space_indent("my block\n of text\nwith indents.")
    >>> print(t)
        my block
         of text
        with indents.
    """
    return '\n'.join(" " * number + t for t in s.splitlines())


_re_newline_space = regex.compile(r'\s*\n\s*')


def stub(s):
    """Takes a string, like a docstring, and returns only the first
    line.

    Parameters
    ----------
    s : str
        The string to generate a stub for.

    Returns
    -------
    stub : str
        The string with its first line stripped.

    Examples
    --------
    >>> stub(stub.__doc__)
    'Takes a string, like a docstring, and returns only the first line.'
    """
    first = s.split('\n\n')[0]
    return _re_newline_space.sub(' ', first,)


_re_multilines = regex.compile(r'(?:\n{2,})')


def strip_multi_newlines(s):
    r"""Strip multiple consecutive newlines in a string.

    Parameters
    ----------
    s : str
        String to strip multiple consecutive newlines from.

    Returns
    -------
    stripped_string : str
        The same string with multiple consecutive newlines replaced to single
        newlines.

    Examples
    --------
    >>> strip_multi_newlines("This is my\nfirst string.")
    'This is my\nfirst string.'
    >>> strip_multi_newlines("This is my\n\nsecond string.")
    'This is my\nsecond string.'
    """
    return _re_multilines.sub("\n", s)


def strip_end_quotes(s):
    """Strip matched quotes from the ends a string.

    Parameters
    ----------
    s : str
        String to strip matched quotes.

    Returns
    -------
    stripped_string : str
        The same string with matched quotes striped.

    Examples
    --------
    >>> strip_end_quotes('This is my "test" string.')
    'This is my "test" string.'
    >>> strip_end_quotes('"This is my test string"')
    'This is my test string'
    >>> strip_end_quotes('"This is my test string "')
    'This is my test string '
    """
    if '\n' in s:
        # Newlines, process each individually
        return '\n'.join(map(strip_end_quotes, s.splitlines()))

    for char in "\'\"":
        if s.count(char) % 2 == 0:  # even number of quotes
            pieces = s.split(char)
            if pieces[0].strip() == '' and pieces[-1].strip() == '':
                return ''.join((pieces[0], char.join(pieces[1:-1]),
                                pieces[-1]))

    return s


_re_find_capitals = regex.compile(r'[A-Z][^A-Z]*')


def convert_camelcase(string, sep='_'):
    """Convert a camel-case string to a string to a lower-case string with
    a seperator character.

    Parameters
    ----------
    string : str
        The camel-case string to convert
    sep : Optional[str]
        The separator character.

    Returns
    -------
    processed_string : str
        The converted string

    Examples
    --------
    >>> convert_camelcase('ProcessContext')
    'process_context'
    >>> convert_camelcase('processcontext')
    'processcontext'
    >>> convert_camelcase('ProcessContext', '.')
    'process.context'
    """
    pieces = _re_find_capitals.findall(string)
    if pieces:
        return sep.join(i.lower() for i in pieces)
    else:
        return string


def find_basestring(strings):
    """Evaluate the common base string amongst a list of strings.

    Parameters
    ----------
    strings : List[str]
        A list of strings

    Returns
    -------
    base_string : str
        The common base string.

    Examples
    --------
    >>> find_basestring(['tester', 'test', 'testest'])
    'test'
    >>> find_basestring(['tester', 'test', 'testest', 'smile'])
    ''
    """
    # Get one of the strings
    base = strings[0]

    # Chop off from the end until a base string that is common to all is found
    base_found = all(i.startswith(base) for i in strings)

    while not base_found and base:
        base = base[:-1]
        base_found = all(i.startswith(base) for i in strings)

    return base


def str_to_list(string):
    """Parse a string into a list.

    Parameters
    ----------
    string : str
        The string with list entries separated by semicolons, newlines or
        commas.

    Returns
    -------
    parsed_list : list
        The parsed list with the string split into string pieces.

    Examples
    --------
    >>> str_to_list('one, two, three')
    ['one', 'two', 'three']
    >>> str_to_list('one, 1; two, 2; three, 3')
    ['one, 1', 'two, 2', 'three, 3']
    >>> str_to_list('''
    ... one, 1
    ... two, 2
    ... three, 3''')
    ['one, 1', 'two, 2', 'three, 3']
    >>> str_to_list("friends")
    ['friends']
    """
    pieces_semicolon = string.split(';')

    if len(pieces_semicolon) > 1:
        return [piece.strip() for piece in pieces_semicolon]

    pieces_newline = [piece.strip() for piece in string.split('\n')]
    pieces_newline = list(filter(bool, pieces_newline))  # remove empty items

    if len(pieces_newline) > 1:
        return pieces_newline

    pieces_comma = string.split(',')

    if len(pieces_comma) > 1:
        return [piece.strip() for piece in pieces_comma]

    return [string]


_re_entry = regex.compile(r'^(?P<space_level>\s*)'
                          r'(?P<key>.+?)'
                          r'(?<!\\)\:'  # match ':' but not '\:'
                          r'\s*'
                          r'(?P<value>.*)')


def str_to_dict(string, strip_quotes=True):
    r"""Parse a string into a dict.

    Parameters
    ----------
    string : str
        The string with dict-like entries separated by colons.
    strip_quotes : Optional[bool]
        If True, matched quotes ('") are stripped at the ends of value strings.

    Returns
    -------
    parsed_dict : dict
        The parsed dict with keys and values as strings.

    Examples
    --------
    >>> string = '''
    ... first: one
    ... second:
    ...   multiline
    ...   entry
    ... third: 3
    ... '''
    >>> d = str_to_dict(string)
    >>> d == {'first': 'one', 'second': '  multiline\n  entry', 'third': '3'}
    True
    """
    d = dict()

    # Split the string into lines and see which lines have an entry of the
    # form "key: value"
    lines = string.splitlines()

    # Go from line to line and identify which lines are key/value pairs

    # Keep a list of all lines that pertain to a given entry
    current_entry_list = []
    # Keep track of how many spaces were used to define the entry. This is used
    # to place sub-blocks of an entry within the entry.
    current_spaces = None

    for line in lines:
        m = _re_entry.match(line)

        # Workup the line. Strip leading spaces from the line.
        if current_spaces is not None:
            line = line[current_spaces:]

        # Replace escaped colons ('\:') with colons (':')
        line = line.replace("\\:", ":")

        if m is not None:
            # If there'string a match, then we may have a new entry.
            # However, only create a new entry if the space_level is the same
            # as the last entry.
            group_dict = m.groupdict()
            space_level = len(group_dict['space_level'])
            key = group_dict['key'].strip()
            value = group_dict['value']

            if current_spaces is not None and space_level > current_spaces:
                # Not a new entry. Just add the line to the last entry.
                current_entry_list.append(line)
            else:
                # A new entry. Create a new current_entry_list.
                current_spaces = space_level
                current_entry_list = d.setdefault(key, [])
                current_entry_list.append(value)
        else:
            # If there'string not a match, just add the line to the current
            # entry
            current_entry_list.append(line)

    # Convert the entries in the dict from a list of strings to strings
    if strip_quotes:
        d = {k: strip_end_quotes("\n".join(v).strip('\n'))
             for k, v in d.items()}
    else:
        d = {k: "\n".join(v).strip('\n') for k, v in d.items()}

    return d


def group_strings(lst):
    """Group adjacent strings in a list and remove empty strings.

    Parameters
    ----------
    lst : Union[list, str or :obj:`Tag <.Tag>`]
        An AST comprising a strings, tags or lists of both.

    Returns
    -------
    processed_list
        The list with the strings grouped and cleaned up

    Examples
    --------
    >>> group_strings(lst=['a', 'b', '', 3, '', 4, 5, 'f'])
    ['ab', 3, 4, 5, 'f']
    >>> group_strings(lst=['a', 'b', 'c', 3, 'd', 'e', 4, 5, 'f'])
    ['abc', 3, 'de', 4, 5, 'f']
    """
    if hasattr(lst, 'content'):
        lst.content = group_strings(lst.content)
    elif isinstance(lst, list):
        # Remove empty strings
        new_list = list(filter(bool, lst))
        lst.clear()
        lst += new_list

        # Join consecutive string elements
        new_list = []
        for cond, group in groupby(lst, key=lambda x: isinstance(x, str)):
            if cond:
                new_list.append(''.join(group))
            else:
                new_list += list(group)
        lst.clear()
        lst += new_list

        # Iterate over the items
        for i, item in enumerate(lst):
            if not isinstance(item, str):
                lst[i] = group_strings(item)

    return lst


_re_macro = regex.compile(r"(?P<macro>" +
                          settings.tag_prefix +  # tag prefix. e.g. '@'
                          r"[\w\.]+)"
                          r"({\s*})?"  # match empty curly brackets
                          )


def replace_macros(s, *dicts):
    """Replace the macros and return a processed string.

    Macros are simple string replacements from context entries whose keys
    start with the tag_prefix character (e.g. '@test'). The
    process_context_asts will ignore macro context entries that have keys
    which start with this prefix. This is to preserve macros as strings.

    Parameters
    ----------
    s : str
        The input string to replace macros within.
    *dicts : Tuple[dict]
        One or more dicts containing variables defined for a specific document.
        Values will be replaced with the first dict found with that value.

    Returns
    -------
    processed_string : str
        A string with the macros replaced.

    Raises
    ------
    MacroNotFound
        Raises a MacroNotFound exception if a macro was included, but it could
        not be found.
    """
    # Replace the values
    def _substitute(m):
        # Get the string for the match
        # ex: macro = '@friend.name'
        d = m.groupdict()
        macro = d['macro']

        # Split at periods
        # ex: pieces = ['@friend', 'name']
        pieces = macro.split('.')

        # See if the first piece corresponds to an entry in kwargs
        obj = None
        while pieces:
            piece = pieces.pop(0)

            if obj is None and any(piece in d for d in dicts):
                for d in dicts:
                    if piece in d:
                        obj = d[piece]
                        break
            elif hasattr(obj, piece):
                obj = getattr(obj, piece)
            else:
                # Match not found. Re-add piece to the pieces list
                pieces.insert(0, piece)
                break

        # Convert obj and the remaining pieces to a string
        if obj is None:
            # no match found. Return the match
            return m.group()
        else:
            # match(es) found, replace with the string
            return str(obj) + ''.join('.' + piece for piece in pieces)

    # Return a string with the dicts substituted. Keep substituting until
    # all dicts are replaced or the string is no longer changing
    s, num_subs = _re_macro.subn(_substitute, s)
    last_num_subs = 0
    while num_subs > 0 and num_subs > last_num_subs:
        last_num_subs = num_subs
        s, num_subs = _re_macro.subn(_substitute, s)

    return s



"""
The 'preview' subcommand.
"""
import click

from .options import in_option, debug_option
from ..server import run_server
from .. import settings


@click.command()
@in_option  # only load the -i flag
@click.option('--port', '-p', show_default=True, default=settings.default_port,
              help="The port to listen to for the webserver")
@debug_option
def preview(in_path, out_dir=None, port=settings.default_port, debug=False):
    """Preview documents with a local webserver"""
    run_server(in_path=in_path, out_dir=out_dir, port=port, debug=debug)



"""
The document context
"""
import weakref
from copy import deepcopy

from .exceptions import TargetNotFound
from ..context import BaseContext
from ..label_manager import LabelManager
from ..signals.signals import signal
from ..paths import SourcePath, TargetPath
from ..utils.string import str_to_list
from ..utils.list import flatten
from .. import settings


find_builder = signal('find_builder')


class DocumentContext(BaseContext):
    """A context dict used for documents objects.
    """

    __slots__ = ()

    #: Required entries in the document context dict to be a valid document
    #: context--as well as their matching type to be checked.
    validation_types = {
        'document': None,
        'environment': None,
        'src_filepath': SourcePath,
        'project_root': SourcePath,
        'target_root': TargetPath,
        'targets': set,
        'paths': list,
        'label_manager': LabelManager,
        'mtime': float,
        'doc_id': str,
        'process_context_tags': set,
        'process_paragraphs': set,
        'label_fmts': dict,
        'label_resets': dict,
        'inactive_tags': set,
    }

    #: The keys for context entries that should not be inherited
    #: from the parent context. They should be unique to each context.
    do_not_inherit = {
        # The document reference in a context should be for the document that
        # owns the context. Consequently, a new context should have a new
        # document and the parent's document should not be inherited.
        'document',

        # A subdocument shouldn't have the same subdocument 'include' entries
        # as the parent. Otherwise, documents could be included multiple times
        # or they may not be found.
        'include', 'includes',

        # Each document should have its own short title
        'short',

        #: The doc_id is unique for each document
        'doc_id',

        # The modification time is the last time a document was changed. This
        # should be different for each document and subdocument. Note that the
        # mtime entry is set by the document object itself so that it can tell
        # when the full context has been read in.
        'mtime',

        # Paths are not inherited since we do not want to use paths for
        # template directories from the parent if the parent's template is
        # different. Also, the paths may be relative to the parent document
        # and may be invalid for a subdocument in a subdirectory.
        'paths',

        # Do not inherit builders. Each document and context has its own
        'builders',

        #: The contents of the body (specified by body_attr) doesn't carry over
        #: because each body has its own body.
        settings.body_attr,
    }

    #: The keys for context entries that should not be removed when the
    #: context is cleared. These are typically for entries setup in the
    #: __init__ of the class.
    exclude_from_reset = {
        # The document that owns the context and its parent variables should
        # not be reset, since we do not want to have this document inherited
        # from the parent.
        'document',

        # The following manager objects should remain the same every time the
        # context is reset. They should persist for a project, and they are
        # defined by the root document and build environment.
        'label_manager',
        'environment',

        # Keep the same list for paths when resetting. Note that will place
        # the entry here to prevent the lists path from being reset by the
        # parent 'reset' method. However, this reset method properly resets the
        # entry in the paths list without creating a new list object.
        'paths'
    }

    # Replace the following mutable entries.
    # When a document or template specified these entries, their values
    # should not be appended to the parent_context values--they should be
    # replaced with the new values.
    replace = {
        'targets',
        'inactive_tags',
    }

    def __init__(self, document, *args, **kwargs):
        self.document = document

        if kwargs.get('parent_context') is None:
            kwargs['parent_context'] = deepcopy(settings.default_context)

        # Conduct the rest of the initializations, including the reset.
        super(DocumentContext, self).__init__(*args, **kwargs)

    def reset(self):
        super(DocumentContext, self).reset()
        # Make sure the following entries are present from the parent or
        # root context
        self.is_valid('environment', 'project_root', 'target_root')

        document = self['document']()
        src_filepath = document.src_filepath
        self['src_filepath'] = src_filepath

        # set the document id
        if self.get('doc_id', None) is None:
            self['doc_id'] = str(document.doc_id)
        else:
            self['doc_id'] = str(self['doc_id'])

        # set the document's mtime
        if src_filepath.exists():
            self['mtime'] = src_filepath.stat().st_mtime

        # The the root document, if it wasn't set already
        if self.get('root_document', None) is None:
            self['root_document'] = weakref.ref(document)
            self['is_root_document'] = True
        else:
            self['is_root_document'] = False

        # Initialize the managers, if this is the root (document) context
        if self.get('label_manager', None) is None:
            self['label_manager'] = LabelManager(root_context=self)

        # Set the document's level. This is based off of the parent context's
        # level
        self['level'] = self.get('level', 0) + 1

        # Prepare the paths entry
        paths = self.setdefault('paths', [])
        paths.clear()

        # Include the path for the src_filepath
        local_path = SourcePath(project_root=src_filepath.project_root,
                                subpath=src_filepath.subpath.parent)
        if local_path not in paths:
            paths.append(local_path)

        # Include the project_root path
        if self['project_root'] not in paths:
            paths.append(self['project_root'])

    @property
    def document(self):
        """The document that owns this context"""
        # Retrieve and de-reference the document
        doc_ref = self.get('document', None)
        return doc_ref() if isinstance(doc_ref, weakref.ref) else None

    @document.setter
    def document(self, value):
        # Create a weakref to the document
        self['document'] = weakref.ref(value)

    @property
    def root_document(self):
        """The root document for a project"""
        # Retrieve and de-reference the document
        doc_ref = self.get('root_document', None)
        return doc_ref() if isinstance(doc_ref, weakref.ref) else None

    @root_document.setter
    def root_document(self, value):
        # Create a weakref to the document
        self['root_document'] = weakref.ref(value)

    @property
    def is_root_document(self):
        return self.get('is_root_document', False)

    @property
    def targets(self):
        """Retrieve a list of targets from the 'targets' or 'target'
        entry of the document.

        Returns
        -------
        target_list : List[str]
            A list of targets specified in the context.
        """
        # Get the targets from the context.
        # In the default context, this is set as the 'targets' entry, which
        # may be over-written by the user. However, a 'target' entry
        # could be
        # used as well (for convenience), and it should be checked first,
        # since
        # it overrides the default 'targets' entry.
        if 'target' in self:
            targets = self['target']
        elif 'targets' in self:
            targets = self['targets']
        else:
            targets = ''

        # Convert to a list, if needed
        target_list = (str_to_list(targets) if isinstance(targets, str) else
                       targets)

        if 'none' in target_list:
            return []

        # Remove empty entries
        target_list = list(filter(bool, target_list))

        # Add trailing period to extensions in target_list, and make sure
        # there are no duplicates
        return {t if t.startswith('.') else '.' + t for t in target_list}

    def target_filepath(self, target):
        """Return the target filepath for this document's target, given
        the target.

        The target filepaths are retrieved from the builders. Accordingly,
        these should already have been loaded in the context. Additionally,
        this function may return target_filepaths for targets not listed by
        the 'targets' property. This is because these targets may be
        intermediary targets stored in a cache directory.

        Parameters
        ----------
        target : str
            The target to search

        Returns
        -------
        target_filepath : :obj:`.paths.TargetPath`
            The filepath for the target file of the document.

        Raises
        ------
        TargetNotFound
            Raised if a target was requested but not found.
        """
        builders = find_builder.emit(context=self, target=target)

        # convert from list of lists to flat list
        builders = list(flatten(builders))

        if len(builders) == 0:
            msg = ("A target_filepath could not be found for the '{}' "
                   "document target")
            raise TargetNotFound(msg.format(target))
        builder = builders[0]
        return builder.outfilepath

    def target_filepaths(self):
        """Return all target filepaths for this document's targets

        The target filepaths are retrieved from the builders. Accordingly,
        these should already have been loaded in the context. Additionally,
        this function may return target_filepaths for targets not listed by
        the 'targets' property. This is because these targets may be
        intermediary targets stored in a cache directory.

        Returns
        -------
        targets : Dict[str, :obj:`.paths.TargetPath`]
            The targets dict. The strings are the targets and the values
            are the corresponding target_filepaths.
        """
        builders = find_builder.emit(context=self)

        # convert from list of lists to flat list
        builders = list(flatten(builders))

        # get the targetfilepaths
        target_filepaths = [builder.outfilepath for builder in builders]

        # Formulate the targets dict. Key names should include a period.
        # ex: '.pdf'
        return {p.suffix: p for p in target_filepaths}

    @property
    def includes(self):
        """Retrieve a list of included subdocument source paths from the
        'include' entry of the context.

        Returns
        -------
        include_list : List[:obj:`SourcePath <.paths.SourcePath>`]
            A list of the paths for the included subdocuments.
        """
        assert self.is_valid('src_filepath')

        # Retrieve the include entry
        if 'include' in self:
            includes = self['include']
        elif 'includes' in self:
            includes = self['includes']
        else:
            includes = None

        # Make sure it's properly formatted as a string for further processing
        if not isinstance(includes, str):
            return []

        # Get the included subdocument paths. Strip extra space on the ends of
        # entries on each line and remove empty entries.
        includes = [path.strip() for path in includes.split('\n')
                    if not path.isspace() and path != '']

        # Reconstruct the paths relative to the context's src_filepath.
        src_filepath = self['src_filepath']
        subpath = src_filepath.subpath.parent
        project_root = src_filepath.project_root

        includes = [SourcePath(project_root=project_root,
                               subpath=subpath / path)
                    for path in includes]
        return includes



"""
The base class for Context objects that include functions to validate a
context.
"""
import logging
from pprint import pprint
from weakref import ref

from .utils import load_from_string
from ..tags import Tag
from ..utils.classes import all_attributes_values
from ..utils.string import str_to_dict, str_to_list
from .. import settings


class ContextException(Exception):
    """An exception raised when processing a BaseContext"""
    pass


class BaseContext(dict):
    """A context dict with entries used for rendering target documents.

    Contexts are suppose to be data container dicts--i.e. there are no
    sophisticated processing or rendering functions. The available functions
    only manage the data and set of the context.

    The BaseContext is basically a heritable dict. It keeps track of dict
    lineage and initial values so that it can be reset to its initialized
    state.

    .. note:: I've tried different implementations, including a ChainMap-like
              inheritance. The problem with these is that key lookup and
              __contains__ lookup are relatively slow, compared to a standard
              dict. For this reason, I've decided to implement the BaseContext
              as a simple dict with values copied from a parent_context. The
              downside of this approach is that it's more memory intensive, but
              this is mitigated to a great extended from creating shallow
              copies or weak references of parent_context entries.

    Parameters
    ----------
    parent_context : Optional[dict]
        A parent or template context.
    *args, **kwargs : tuple and dict
        The entries to populate in the context dict.

    Attributes
    ----------
    validation_types : Dict[str, type]
        A listing of entry keys and the types they should be.
        If None is listed as the type, then a type check will not be conducted.
    parent : :obj:`BaseContext <.context.BaseContext>`
        A weakref to a BaseContext of a parent document.
    do_not_inherit : Set[str]
        The context keys that should not be accessed (inherited) from the
        parent context.
    exclude_from_reset : Set[str]
        The context entries that should not be removed when the context is
        reset.
    replace : Set[str]
        The *muttable* context entries that should be replaced, rather than
        appended, from the parent context.
    _initial_values : dict
        A dict containing the initial values. Since this starts with an
        underscore, it is hidden when listing keys with the keys() function.
    _parent_context : :obj:`Type[BaseContext] <.BaseContext>`
        A dict containing the parent context from which values may be
        inherited. Since this starts with an underscore, it is hidden when
        listing keys with the keys() function.

    Examples
    --------
    >>> parent = BaseContext(a=1, b=[])
    >>> child = BaseContext(parent_context=parent)
    >>> child['a']
    1
    >>> child['a'] = 2
    >>> child['a']
    2
    >>> parent['a']
    1
    >>> child['b'].append(1)
    >>> child['b']
    [1]
    >>> parent['b']
    []
    """

    # A __dict__ attribute would be redundant
    __slots__ = ('__weakref__', '_parent_context', 'initial_values')

    validation_types = dict()

    do_not_inherit = set()
    exclude_from_reset = set()
    replace = set()

    _do_not_inherit = None

    def __init__(self, *args, **kwargs):
        super().__init__()

        kwargs.update(*args)

        # Remove protected keys
        for k in {'_parent_context', '_initial_values'}:
            if k in kwargs:
                del kwargs[k]

        # Store the parent context
        if 'parent_context' in kwargs:
            parent = kwargs.get('parent_context', None)
            del kwargs['parent_context']
            self.parent_context = parent

        # Store the initial values
        initial_values = dict(*args, **kwargs)
        self.initial_values = initial_values

        # Reset the dict with the default_context and parent_context values
        self.reset()

    def __repr__(self):
        return (self.__class__.__name__ + '{' +
                ', '.join('{}: {}'.format(k, self[k])
                          for k in self.keys()) +
                '}')

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        # Copy over the entries for this context dict
        result.update(self)

        # Copy over the parent_context and initial_values
        result.parent_context = self.parent_context
        result.initial_values = self.initial_values

        return result

    @classmethod
    def find_do_not_inherit(cls):
        """Retrieve a union set for the do_not_inherit attribute of this class
        and all parent classes.

        Returns
        -------
        do_not_inherit : Set[str]
            A set of all attributes that should not be inherited by child class
            instances.
        """
        if cls.__dict__.get('_do_not_inherit') is None:
            do_not_inherit = all_attributes_values(cls=cls,
                                                   attribute='do_not_inherit')
            cls._do_not_inherit = set(do_not_inherit)
        return cls._do_not_inherit

    @property
    def parent_context(self):
        """The parent context."""
        # Try to get the keys from the parent context
        parent = getattr(self, '_parent_context', None)

        # dereference the parent, if needed
        if parent is not None:
            parent = parent() if isinstance(parent, ref) else parent

        return parent

    @parent_context.setter
    def parent_context(self, parent):
        # Set the parent_context in self's dict. Create a weakref, if able
        # to do so, because this context does not own the parent_context
        self._parent_context = (ref(parent)
                                if hasattr(parent, '__weakref__') else
                                parent)

    def load(self, string, strip_header=True, overwrite=True):
        """Load context entries from a string.

        The load function interprets a string containing a dict in the format
        processed by :func:`str_to_dict
        <disseminate.utils.string.str_to_dict>`. Additionally, this function
        parses header portions of strings delineated by 3 or more '-'
        characters, and only new entries entries that match the types of
        existing entries are inserted.

        Parameters
        ----------
        string : str
            The string to load into this context.
        strip_header : Optional[bool]
            If True (default), the returned string will have the header
            removed.
        overwrite : Optional[bool]
            If True, allow existing entries in self to be overwritten by
            changes.

        Returns
        -------
        string : str
            The processed string with the header removed (if available) or just
            the string passed as an argument.

        Examples
        --------
        >>> context = BaseContext()
        >>> rest = context.load('test: 1')
        >>> print(context)
        BaseContext{test: 1}
        """
        # Process the header block into a dict
        rest, d = load_from_string(string)

        # update self with the dict entries
        self.match_update(changes=d, overwrite=overwrite)

        return rest if rest is not None and strip_header else string

    def reset(self):
        """(Selectively) resets the context to its initial state.

        The context is reset by removing items with keys not specified in the
        'exclude_from_clear' class attribute.

        .. note:: Entries in the parent_context are copied to this context.
                  If the parent_context entry is a mutable, like a list or
                  dict, then a copy of that mutable is created if a 'copy'
                  method exists for that mutable. However, mutables within
                  those mutables will still point to the original.

        Examples
        --------
        >>> l1, l2 = ['a'], ['b']
        >>> context = BaseContext(l1=l1)
        >>> context['l2'] = l2
        >>> context['l1'].append(1)
        >>> context['l2'].append(2)
        >>> context.reset()
        >>> context['l1'] == ['a']  # reset to original l1
        True
        >>> 'l2' in context
        False
        """
        # 1. The first step is to remove all entries except those listed in
        #    the 'exclude_from_reset' attributes.

        # Get a set of keys to exclude from reset from this class and
        # subclasses, and exclude these from the keys to remove
        exclude = all_attributes_values(cls=self.__class__,
                                        attribute='exclude_from_reset')
        keys_to_remove = self.keys() - exclude

        # Now remove the entries that are remaining
        for k in keys_to_remove:
            del self[k]

        # Copy in the parent value arguments, except those that shouldn't
        # be inherited
        parent_context = self.parent_context
        if parent_context is not None:
            do_not_inherit = self.find_do_not_inherit()
            keys_to_inherit = parent_context.keys() - do_not_inherit
            self.match_update(parent_context, keys=keys_to_inherit)

        # Copy in the initial value arguments. The initial values should
        # not be modified, so copies of mutabes are created
        self.match_update(changes=self.initial_values)

    def is_valid(self, *keys, must_exist=True):
        """Validate the entries in the context dict.

        This function checks that the keys denoted by the validate_types
        class attribute match the corresponding types for its values.

        Parameters
        ----------
        keys : Optional[Tuple[str]]
            If specified, only the given keys will be checked if they're in
            the validate_types class attribute. Otherwise all validate_types
            will be checked.
        must_exist : Optional[bool]
            If True (default), then the entry must also exist in addition to
            having the correct type.
            If False, then entries can be missing in this dict that are listed
            in the validate_types class attribute.

        Returns
        -------
        validated : bool
            True if the context is valid.
            False if the context is not valid
        """

        # Determine which keys to use. Use all validation keys, if not keys
        # were given
        keys = self.validation_types.keys() if len(keys) == 0 else keys

        for key in keys:
            if key not in self:
                msg = "The key '{}' could not be found in the context."
                logging.debug(msg.format(key))

                if must_exist:
                    return False
                continue

            value_type = self.validation_types.get(key, None)

            if value_type is not None:
                valid_entry = isinstance(self[key], value_type)
                if value_type is not None and not valid_entry:
                    msg = ("The key '{}' has a value '{}' that is not valid; "
                           "a value type of '{}' is expected.")
                    logging.debug(msg.format(key, self[key], value_type))

                    return False

        return True

    def match_update(self, changes, keys=None, overwrite=True, level=1):
        """Update this context dict with changes by matching entry types.

        Matched update behaves like a dict's update method with these key
        differences:

        1. If changes is a string, it is first converted to a dict.
        2. Existing entries from changes are converted to the same type as the
           value in this dict. See examples below.
        3. Mutable values are deep copied from the changes dict.

        Parameters
        ----------
        changes : Union[str, dict, :obj:`.BaseContext`]
            The changes to update into this context dict.
        keys : Optional[Iterable[str]]
            If specified, only update the given keys
        overwrite : Optional[bool]
            If True, allow existing entries in self to be overwritten by
            changes.
        level : Optional[int]
            The level of recursion for this function.

        Notes
        -----
        The matched update only updates entries for matching types and appends
        to nested mutables, like lists and dicts.

        1. New entries from 'changes' are copied

           a. If the change's value has a 'copy' method, a copy is created
              with that method
           b. Otherwise, the value is copied directly

        2. Existing entries are converted to the type of the value in this
           context dict.

           a. Lists, sets and dicts: A copy of the changes value is converted
              to the respective type and appended to this context dict's list,
              set or dict.
           b. Immutables are converted and added directly.

        Examples
        --------
        >>> orig = BaseContext(test = [])
        >>> orig.match_update(changes={'test': 'test'})
        >>> orig['test']
        ['test']
        >>> orig.match_update(changes={'test': 'more changes'})
        >>> orig['test']
        ['more changes', 'test']
        """
        # When first invoking this function, 'changes' may simply be a string
        # in YAML format. If so, convert it to a dict to pull values from
        if level == 1 and isinstance(changes, str):
            changes = str_to_dict(changes)

        # Make sure the context isn't too deeply nested
        if level >= settings.context_max_depth:
            msg = "Context cannot exceed a depth of {}."
            raise ContextException(msg.format(settings.context_max_depth))

        # Select the keys for entries to copy over
        keys = changes.keys() if keys is None else keys

        # Copy the entries in the changes dict one-by-one
        for key in keys:
            if key not in changes:
                # Skip if the key is not found in changes
                continue

            change_value = changes[key]

            # 1. *Missing entries*. Copy values that are not in the
            #    original--whether the key is actually in the original or in
            #    the parent_context
            if key not in self.keys():
                try:
                    # Try making a copy. Some objects, like tags, take a
                    # new_context parameter to set the new context in the
                    # created object
                    self[key] = change_value.copy(new_context=self)

                except TypeError:
                    # Other objects, like dicts, have a copy method to create
                    # a shallow copy, but these do not accept keyword arguments
                    self[key] = change_value.copy()
                except AttributeError:
                    # Otherwise reference the value directly, like immutables
                    self[key] = change_value
                continue

            # At this point, the key already exists in self. If overwrite is
            # not enabled, then don't do anything else (i.e. don't overwrite)
            if not overwrite:
                continue

            # Now copy over changed values based on the type of the original
            # value's type
            original_value = self[key]

            # Clear the original_value if this value should be replaced
            if key in getattr(self, 'replace', set()):
                try:
                    original_value.clear()
                except AttributeError:
                    msg = ("The context entry '{}' is of type '{}', and it "
                           "cannot be cleared for replacement.")
                    logging.warning(msg.format(original_value,
                                               type(original_value)))

            # 2. *lists*. Create a copy from changes
            if isinstance(original_value, list):
                change_value = (str_to_list(change_value)
                                if isinstance(change_value, str) else
                                list(change_value))
                original_value[0:0] = change_value  # prepend to top

            # 3. *sets*. Create a copy from changes
            elif isinstance(original_value, set):
                change_value = (str_to_list(change_value)
                                if isinstance(change_value, str) else
                                list(change_value))
                original_value |= set(change_value)

            # 4. *dicts*. Create a copy from changes
            elif isinstance(original_value, dict):
                change_value = (str_to_dict(change_value)
                                if isinstance(change_value, str) else
                                dict(change_value))
                BaseContext.match_update(self=original_value,
                                         changes=change_value,
                                         level=level + 1)

            # 5. *tags*. Create a copy
            elif isinstance(original_value, Tag):
                if isinstance(change_value, str):
                    # If the change_value is a string, then convert it to
                    # a tag of the same type
                    tag_cls = original_value.__class__
                    tag = tag_cls(name=key, content=change_value,
                                  attributes='', context=self)
                elif isinstance(change_value, Tag):
                    tag = change_value.copy(new_context=self)
                else:
                    # Do nothing if the change value is neither a tag or a
                    # string, as this cannot be converted
                    continue
                self[key] = tag

            # For immutable types, like ints, covert strings into their
            # proper format and replace the original's value
            elif isinstance(original_value, bool):
                change_value_lower = change_value.lower().strip()
                if change_value_lower == 'false':
                    self[key] = False
                elif change_value_lower == 'true':
                    self[key] = True

            elif isinstance(original_value, int):
                try:
                    self[key] = int(change_value)
                except ValueError:
                    pass

            elif isinstance(original_value, float):
                try:
                    self[key] = float(change_value)
                except ValueError:
                    pass

            # Otherwise, if the original value's type and change value type
            # match, like for strings, just replace the original's value
            elif type(original_value) == type(change_value):
                self[key] = change_value

    def filter(self, keys):
        """Create a new context with the entries copied for the given keys.

        Parameters
        ----------
        keys : Iterable[str]
            The keys for entries to include in the returned filtered
            BaseContext

        Returns
        -------
        copy : :obj:`.context.BaseContext`
            The filtered copy of the BaseContext.
        """
        # Create an empty copy
        cls = self.__class__
        copy = cls.__new__(cls)

        # Update the copy with the values specified by keys
        copy.match_update(changes=self, keys=keys)
        return copy

    def print(self):
        """Pretty print this context"""
        pprint(self)



"""
Calculate the hash for a tag's contents.
"""

from ..signals import tag_created
from ...utils.string import hashtxt


@tag_created.connect_via(order=50)
def process_hash(tag, tag_factory, **kwargs):
    """A receiver to create a hash for the contents of tags."""
    content = tag.content
    if isinstance(content, str):
        tag.hash = hashtxt(tag.content)
    return tag



"""
A scanner object to find implicit dependencies.
"""
from ...paths import SourcePath, TargetPath
from ...utils.classes import all_subclasses


class Scanner(object):
    """A scanner object parses the contents of a file and finds implicit file
    dependencies.
    """

    extensions = tuple()

    _scanners = None

    def __init__(self):
        msg = "Scanner classes cannot be instantiated"
        raise TypeError(msg)

    @staticmethod
    def scan_function(content):
        """The function to scan content for new infilepaths.

        Subclasses derive this function."""
        return []

    @classmethod
    def get_scanner(cls, extension):
        """Get the scanner for the given extension"""
        if cls._scanners is None:
            # Get all subclasses
            scanners = all_subclasses(Scanner)

            # Create concrete scanners
            scanners_dict = {ext: scanner for scanner in scanners
                             for ext in scanner.extensions}
            cls._scanners = scanners_dict
        return cls._scanners.get(extension, None)

    @classmethod
    def scan(cls, parameters, raise_error=True):
        """Scan parameters for dependencies.

        Parameters
        ----------
        str_or_filepath : Union[str, :obj:`pathlib.Path`]
            A string or a filepath to a filename to scan.
        paths : List[:obj:`pathlib.Path`]
            A list of directories to search.
        raise_error : bool
            If True (default), raise a FileNotFoundError if a dependency
            file could not be found.

        Returns
        -------
        new_infilepaths : List[:obj:`.paths.SourcePath`]
            A list of infilepath dependencies.
        """
        # Prepare the arguments
        parameters = (parameters if isinstance(parameters, list) or
                      isinstance(parameters, tuple) else [parameters])

        new_infilepaths = []

        # Parse each filepath
        for parameter in parameters:
            if (not isinstance(parameter, SourcePath) and
               not isinstance(parameter, TargetPath)):
                continue

            # Make sure this scanner can deal with this type of infilepath
            if parameter.suffix not in cls.extensions:
                # See if an alternative scanner can be found
                scanner = cls.get_scanner(extension=parameter.suffix)
                if scanner is None:
                    # If not, skip this infilepath
                    continue
                else:
                    # If so, use it instead
                    new_infilepaths += scanner.scan(parameter,
                                                    raise_error=raise_error)
                    continue

            # Parse the infilepath, if it exists
            if not parameter.is_file():
                continue
            stubs = cls.scan_function(parameter.read_text())

            # Find the new, valid infilepaths. First, parse the source
            # infilepath.
            if getattr(parameter, 'project_root', None) is not None:
                root = parameter.project_root
            if getattr(parameter, 'target_root', None) is not None:
                root = parameter.target_root
            if getattr(parameter, 'target', None) is not None:
                root = root / parameter.target
            subpath = parameter.subpath.parent

            for stub in stubs:
                # Strip leading slashes so that the stub is not an absolute
                # path
                stub = stub.strip('/')

                test_paths = [SourcePath(project_root=root,
                                         subpath=subpath / stub),
                              SourcePath(project_root=root,
                                         subpath=stub)]
                valid_paths = [p for p in test_paths if p.is_file()]

                # If no files are found, raise an exception
                if not valid_paths and raise_error:
                    msg = "Could not find the file '{}' in paths: '{}'"
                    raise FileNotFoundError(msg.format(stub, test_paths))
                else:
                    new_infilepaths.append(valid_paths[0])

        return new_infilepaths



"""
Builders for EPUB files
"""
import zipfile
import pathlib
import uuid
import logging
import os.path
from datetime import datetime, timezone

from .composite_builders import SequentialBuilder
from ..paths.utils import find_file
from ..utils.classes import weakattr
from ..utils.string import slugify


class XHtml2Epub(SequentialBuilder):
    """Convert xhtml files to an epub v3 file.

    Parameters
    ----------
    parameters, args : Tuple[:obj:`pathlib.Path`, str, tuple, list]
        The input parameters (dependencies), including filepaths for xhtml
        files, for the build. The parameters must include a 'toc.xhtml' file
        for the Table of Contents.
    """

    available = True
    priority = 1000
    action = 'Build xhtml2epub'

    infilepath_ext = '.xhtml'
    outfilepath_ext = '.epub'
    scan_parameters_on_init = False

    context = weakattr()
    package_subdir = pathlib.Path('xhtml')  # dir in ebook to store content
    toc_filename = 'toc.xhtml'

    _render_context = None

    def __init__(self, env, context, **kwargs):
        # Setup the arguments
        assert context is not None

        super().__init__(env, **kwargs)
        self.context = context

        # Check the necessary parameters
        # One of the file should have a toc.xhtml
        assert any(p.name == self.toc_filename for p in self.parameters
                   if hasattr(p, 'name')), "epub requires a toc.xhtml file"

        # Create a subbuilder for the content.opf file
        opf_builder = self.create_opf_builder()
        self.subbuilders.append(opf_builder)

    @property
    def parameters(self):
        parameters = SequentialBuilder.parameters.fget(self)

        # The parameters may included multiple references to the same files,
        # like css files, when building a tree. Remove these
        filtered_parameters = []
        resolved_parameters = []
        for parameter in parameters:
            if not hasattr(parameter, 'resolve'):
                filtered_parameters.append(parameter)
                continue

            resolved_parameter = parameter.resolve()
            if resolved_parameter not in resolved_parameters:
                resolved_parameters.append(resolved_parameter)
                filtered_parameters.append(parameter)

        return filtered_parameters

    @parameters.setter
    def parameters(self, value):
        SequentialBuilder.parameters.fset(self, value)

    def create_opf_builder(self, template_name='default/xhtml/content.opf'):
        """Create a render builder for the content.opf file.

        Parameters
        ----------
        template_name : Optional[str]
            The name of the template to use for the content.opf.
        """
        # Find the render builder class
        builder_cls = self.find_builder_cls(in_ext='.render')

        # Setup a renderer for the content.opfs
        fps = [p for p in self.parameters if hasattr(p, 'suffix')]

        # Place the title.xhtml and toc.xhtml file at the front of the
        # filepaths
        front_fps = ([fp for fp in fps if fp.name == 'title.xhtml'] +
                     [fp for fp in fps if fp.name == 'toc.xhtml'])
        fps = front_fps + [fp for fp in fps if fp not in front_fps]

        render_context = (self._render_context or
                          self.context.filter(['paths', 'title', 'date',
                                               'epub']))
        render_context['template'] = template_name
        render_context['xhtml_files'] = self.filepath_dict_list(fps, '.xhtml')
        render_context['svg_files'] = self.filepath_dict_list(fps, '.svg')
        render_context['css_files'] = self.filepath_dict_list(fps, '.css')

        # Add required entries, if not present
        now = datetime.now(timezone.utc).replace(microsecond=0)
        date_str = now.isoformat().replace("+00:00", "Z")
        render_context.setdefault('uuid', str(uuid.uuid4()).upper())
        render_context.setdefault('title', 'Default Title')
        render_context.setdefault('date', date_str)

        self._render_context = render_context

        # Set the outfilepath for the content.opf
        outfilepath = self.outfilepath
        outfilepath = outfilepath.use_subpath('content.opf')

        # Create the builder
        builder = builder_cls(env=self.env, context=render_context,
                              outfilepath=outfilepath, render_ext='.opf',
                              use_cache=True)
        return builder

    def filepath_dict_list(self, filepaths, suffix):
        """Create a list of dicts from the given filepaths"""
        rv = []
        for filepath in filepaths:
            if filepath.suffix != suffix:
                continue

            # Is the filepath a toc file?
            is_toc = filepath.name == self.toc_filename

            # Generate the subpath for the file, as this is the relative
            # path in the epub container. Make sure to strip '..' relpath
            # elements, as this can create compatibility issues.
            subpath = str(os.path.normpath(filepath.subpath))

            # Generate the navigation identifier for the file
            id = slugify(subpath) if not is_toc else 'toc'

            rv.append({'id': id,
                       'subpath': subpath,
                       'is_toc': is_toc})
        return rv

    def build(self, complete=False):
        # Scan for additional dependencies. This is done during the build,
        # rather during __init__ (as is usually done), because the xhtml
        # files might not have been created until a previous builder has
        # built them
        self.scan_parameters()

        # Build subbuilders first
        super().build(complete=complete)

        context = self.context
        outfilepath = self.outfilepath
        root_path = pathlib.Path('.')  # Base path for zip file

        logging.debug("Creating epub file '{}'".format(outfilepath))

        # Create the epub
        with zipfile.ZipFile(outfilepath, 'w',
                             compression=zipfile.ZIP_DEFLATED) as epub:
            # Add the mimetype file
            epub.writestr('mimetype',
                          "application/epub+zip".encode(),
                          zipfile.ZIP_STORED)

            # Find and add the container.xml file
            container_filepath = (pathlib.Path('xhtml') / 'META-INF' /
                                  'container.xml')
            contain_file = find_file(container_filepath, context=context)

            epub.write(contain_file,
                       arcname=root_path / 'META-INF' / 'container.xml')

            # Create a builder for the content.opf
            builder = self.create_opf_builder()
            assert builder.build(complete=True)
            arc_name = self.package_subdir / builder.outfilepath.subpath
            epub.write(builder.outfilepath, arcname=arc_name)

            # Add files listed in the parameters
            filepaths = self.infilepaths

            for filepath in filepaths:
                # Rework the filepath for the archive. Use the subpath, if
                # available
                arcname = (filepath.subpath
                           if hasattr(filepath, 'subpath') else filepath)

                # Place the content in the package subdir (ex: 'xhtml/')
                arcname = self.package_subdir / arcname

                # Write the content file to the zip file.
                epub.write(filepath, arcname=arcname)

        self.build_needed(reset=True)
        return 'done'



"""
A receiver to load the document's string in the document context
"""
from ..signals import document_onload
from ... import settings


@document_onload.connect_via(order=200)
def load_document(document, **kwargs):
    """Load the document text file into the document context."""
    # Load the string from the src_filepath,
    string = document.src_filepath.read_text()

    # Place the text of the string in the 'body' attribute of the
    # context (see settings.body_attr)
    body_attr = settings.body_attr
    document.context[body_attr] = string
    return document



"""
Tags for data sources
"""
import abc
from io import StringIO

import pandas as pd

from .tag import Tag
from .utils import format_content
from ..paths.utils import find_files
from ..formats.xhtml import xhtml_tag
from ..formats.tex import tex_cmd


class Cell(Tag):
    """A cell in a table"""

    active = True
    html_name = 'td'

    def tex_fmt(self, **kwargs):
        tex = super().tex_fmt(**kwargs)
        tex = tex.replace('&', '\\&')
        return tex


class HeaderCell(Tag):
    """A header cell in a table"""

    active = True
    html_name = 'th'

    def tex_fmt(self, **kwargs):
        tex = super().tex_fmt(**kwargs)
        tex = tex.replace('&', '\\&')
        return tex


class Data(Tag):
    """A container for data.

    Attributes
    ----------
    dataframe : :obj:`pandas.DataFrame`
        The raw data parsed in a dataframe
    processed_headers : Union[List[:obj:`tags.Tag`], None]
        The headers processed into tags
    processed_rows : List[List[:obj:`tags.tag`]]
        The rows processed into tags
    """

    active = False
    process_content = False

    dataframe = None
    processed_headers = None
    processed_rows = None

    def __init__(self, name, content, attributes, context):
        super().__init__(name=name, content=content, attributes=attributes,
                         context=context)
        # See if the content containts filenames or the actual data
        filepaths = find_files(string=content, context=context)

        if filepaths:
            # If it's a filepath, send that to the load function
            self.load(filepaths[0])
        else:
            # Otherwise consider that its data. Load that.
            self.load(StringIO(content))

    @abc.abstractmethod
    def load(self, filepath_or_buffer):
        """Load the dataframe from a filepath or a string buffer.

        Parameters
        ----------
        filepath_or_buffer : Union[:obj:`pathlib.Path`, str, \
        :obj:`io.StringIO`]
            The filename and path (filepath) or string buffer.
        """
        pass

    def process_tags(self):
        """Process the tags for the items in the dataframe."""
        assert self.dataframe is not None, "Pandas dataframe not loaded."

        # Process the headers, if needed
        headers = self.headers
        if self.processed_headers is None and headers:
            hdrs = [HeaderCell(name='cell', content=str(header), attributes='',
                               context=self.context) for header in headers]
            self.processed_headers = hdrs

        # Process the rows, if needed
        if self.processed_rows is None:
            rows = []
            for row in self.rows:
                parsed = [Cell(name='cell', content=str(i), attributes='',
                               context=self.context)
                          for i in row[1:]]
                rows.append((row[0],) + tuple(parsed))
            self.processed_rows = rows

    @property
    def headers(self):
        """The list of the data headers"""
        return (None if 'noheader' in self.attributes else
                list(self.dataframe.columns))

    @property
    def rows(self):
        """An iterator for the data rows"""
        return self.dataframe.itertuples()

    @property
    def num_cols(self):
        """The number of columns in the data"""
        columns = getattr(self.dataframe, 'columns', None)
        return len(columns) if columns is not None else None

    @property
    def num_rows(self):
        """The number of columns in the data"""
        rows = getattr(self.dataframe, 'columns', None)
        return len(rows) if rows is not None else None


class DelimData(Data):
    """A container for delimiter data (ex: comma-separated values)"""

    active = True
    aliases = ('csv', 'tsv')
    delimiter = None

    def __init__(self, delimiter=',', *args, **kwargs):
        self.delimiter = delimiter
        super().__init__(*args, **kwargs)

    def load(self, filepath_or_buffer, delimiter=None):
        if self.dataframe is None:
            delimiter = delimiter if delimiter is not None else self.delimiter
            if 'noheader' in self.attributes:
                self.dataframe = pd.read_csv(filepath_or_buffer, engine='c',
                                             header=None,
                                             skipinitialspace=True,
                                             delimiter=delimiter)
            else:
                self.dataframe = pd.read_csv(filepath_or_buffer, engine='c',
                                             skipinitialspace=True,
                                             delimiter=delimiter)

    def tex_table(self, content=None, attributes=None, mathmode=False,
                  level=1):
        # Load the tags
        self.process_tags()

        headers = self.processed_headers
        tex = tex_cmd('toprule') + "\n"

        if headers is not None:
            tex += " & ".join([format_content(cell, 'tex_fmt',
                                              mathmode=mathmode, level=level)
                               for cell in headers]) + " \\\\\n"
            tex += tex_cmd('midrule') + "\n"

        rows = []
        for row in self.processed_rows:
            str = " & ".join([format_content(cell, 'tex_fmt', level=level,
                                                   mathmode=mathmode)
                              for cell in row[1:]]) + " \\\\\n"
            rows.append(str)
        tex += "".join(rows)

        tex += tex_cmd('bottomrule')
        return tex

    def html_table(self, format_func='html_fmt', method='html', level=1,
                   **kwargs):
        # Load the tags
        self.process_tags()

        headers = self.processed_headers

        # Prepare the header row, if a header is available
        elements = []
        if headers is not None:
            header_row = [format_content(cell, format_func, level=level)
                          for cell in headers]

            tr = xhtml_tag('tr', formatted_content=header_row, method=method,
                           level=level)
            thead = xhtml_tag('thead', formatted_content=tr, method=method,
                              level=level)
            elements.append(thead)

        # Prepare each row individually. Each row is a named tuple with the
        # first element as the index
        rows = []
        for row in self.processed_rows:
            body_row = [format_content(cell, format_func, level=level)
                        for cell in row[1:]]

            tr = xhtml_tag('tr', formatted_content=body_row, method=method,
                           level=level)
            rows.append(tr)

        tbody = xhtml_tag('tbody', formatted_content=rows, method=method,
                          level=level)
        elements.append(tbody)

        return elements



"""
Functions to load projects in a session.
"""
import logging

from .store import store
from ...builders.environment import Environment


def load_projects(app):
    """Retrieve the root documents from the store.
    Returns
    -------
    root_documents : List[:obj:`Document <.Document`]
        The loaded root documents.
    """
    # Get the session and config
    settings = app.settings

    # Make sure the project list is loaded
    if 'root_documents' not in store:
        # Get project_filenames
        in_path = settings.get('in_path', '')

        # Fetch the root documents
        envs = Environment.create_environments(root_path=in_path)
        docs = [env.root_document for env in envs]

        # Log the loaded root documents
        for doc in docs:
            logging.debug("Loaded root document '{}'".format(doc))

        # Store the root documents in the global store
        store['root_documents'] = docs

    # See if any of the docs need to be built
    [doc.build() for doc in store['root_documents']]
    return store['root_documents']



"""
Utilities for checkers.
"""
import re
import operator


def name_and_version(string):
    """Parse the name and version from a package name string with an optional
    specifier.

    Parameters
    ----------
    string : str
        The string with a name and version to parase.

    Returns
    -------
    name, operator, version : Union[None, \
        Tuple[str, Union[operator, None], Union[None, Tuple[int]]]
        The name, operator and version for a package.
        None, if the string couldn't be parsed.

    Examples
    --------
    >>> name_and_version('test')
    ('test', None, None)
    >>> name_and_version('test-ab_one')
    ('test-ab_one', None, None)
    >>> name_and_version('test-ab_one>=0.3.1')
    ('test-ab_one', <built-in function ge>, (0, 3, 1))
    >>> name_and_version('new-package>1.2')
    ('new-package', <built-in function gt>, (1, 2))
    >>> name_and_version('jinja2>=2.10')
    ('jinja2', <built-in function ge>, (2, 10))
    """

    m = re.search(r'([\w\.\-\_]+)'  # match package name
                  r'\s*(==|>=|<=|>|<|!=)?\s*'  # operator
                  r'([\d\.]+)?',  # version number
                  string)

    if m is None:
        # The string couldn't be parse. Return None
        return None

    name = m.group(1)
    op = m.group(2)
    version = m.group(3)

    # Convert the operator
    if op is not None:
        op_mapping = {'==': operator.eq,
                      '>': operator.gt,
                      '<': operator.lt,
                      '>=': operator.ge,
                      '<=': operator.le,
                      '!=': operator.ne}
        op = op_mapping[op]

    # Convert the version to a tuple of integers
    if version is not None:
        version = tuple(map(int, version.split('.')))

    return name, op, version




"""
Tags to render asymptote (asy) figures and diagrams
"""
from .img import Img


class Asy(Img):
    """The asy tag for inserting asymptote images."""

    active = True
    process_content = False
    in_ext = '.save'



from ..builder import Builder


class CompositeBuilder(Builder):
    """A builder that integrates multiple (sub)-builders

    Parameters
    ----------
    subbuilders : Optional[List[:obj:`.builders.Builder`]
        The subbuilders to run as part of this composite builder.

    Attributes
    ----------
    clear_done : bool
        If True (default), remove 'done' subbuilders during the build.
    """
    subbuilders = None

    active_requirements = ('priority',)
    priority = 1000
    parallel = False
    clear_done = True

    def __init__(self, env, subbuilders=None, **kwargs):
        super().__init__(env, **kwargs)

        # Load the subbuilders
        subbuilders = (list(subbuilders) if isinstance(subbuilders, list) or
                       isinstance(subbuilders, tuple) else [])
        subbuilders = [sb for sb in subbuilders if isinstance(sb, Builder)]
        self.subbuilders = subbuilders

    def __len__(self):
        number_subbuilders = 0
        for subbuilder in self.subbuilders:
            number_subbuilders += 1
            if hasattr(subbuilder, 'subbuilders'):
                number_subbuilders += len(subbuilder.subbuilders)
        return number_subbuilders

    @property
    def status(self):
        statuses = [sb.status for sb in self.subbuilders]
        statuses = [status for status in statuses if status != 'done']

        # Return 'done' if all statuses are done (or there are no subbuilders)
        # Otherwise return the first non-done status
        return 'done' if len(statuses) == 0 else statuses[0]

    def run_cmd_args(self):
        """Format the for all sub commands

        Returns
        -------
        run_cmd_args : Tuple[str]
            A tuple of the arguments for all sub-builders
        """
        args = []
        for subbuilder in self.subbuilders:
            args += list(subbuilder.run_cmd_args())
        return tuple(args)

    def build(self, complete=False):
        def run_build_once(self):
            """Execute one round of the subbuilders"""
            status = 'done'

            for builder in self.subbuilders:
                status = builder.status

                if status == 'done':
                    # The builder is responsible for resetting its build_needed
                    # method after the build is done.
                    continue

                elif status == 'building':
                    if self.parallel:
                        continue
                    else:
                        break

                elif status == 'ready':
                    builder.build()
                    status = 'building'
                    if self.parallel:
                        continue
                    else:
                        break

                else:
                    return status

            return status

        if complete:
            while self.status in {'building', 'ready'}:
                run_build_once(self)
        else:
            if self.status in {'building', 'ready'}:
                run_build_once(self)

        # Get the current status and reset build_needed states
        status = self.status
        if status == 'done':
            self.build_needed(reset=True)

        # Remove finished builders, if specified by 'clear_done'
        if self.clear_done:
            for subbuilder in list(self.subbuilders):
                if subbuilder.status == 'done':
                    self.subbuilders.remove(subbuilder)

        return status

    def flatten(self, builder=None):
        """Generate a flat list with this builder and all subbuilders
        (recursively).

        Parameters
        ----------
        builder : Optional[:obj:` <.composite_builders.CompositeBuilder>`]
            If specified, flatten the given builder, instead of this builder.

        Returns
        -------
        flattened_list : List[:obj:` <.composite_builders.CompositeBuilder>`]
            The flattened list of builders.
        """
        builder = builder if builder is not None else self
        flattened_list = [builder]  # add the given tag to the list

        # Traverse the items and process each
        for subbuilder in self.subbuilders:
            # Add the item to the list
            if isinstance(subbuilder, CompositeBuilder):
                flattened_list += subbuilder.flatten(builder=subbuilder)
            else:
                flattened_list.append(subbuilder)

        return flattened_list

    def futures(self, builder=None):
        """Generate a list of this builder's future and all sub-builder futures
        (recursively)

        Parameters
        ----------
        builder : Optional[:obj:` <.composite_builders.CompositeBuilder>`]
            If specified, flatten the given builder, instead of this builder.

        Returns
        -------
        futures : List[:obj:` <concurrent.futures.Future>`]
            The flattened list of futures.
        """
        builder = builder if builder is not None else self
        return [b.future for b in self.flatten(builder=builder)
                if getattr(b, 'future', None) is not None]

    def print(self, level=1, max_level=None):
        """Print the builder and subbuilders"""

        def print_builder(b, level, num_spaces=2):
            msg = "  " * num_spaces * level
            msg += str(b)  # the __repr__ of the builder
            if not b.missing_parameters:
                msg += "({}) ".format(b.outfilepath.subpath)
            else:
                msg += " "

            # Get builder attributes that are useful to print out
            attrs = ["{}={}".format(attr, getattr(b, attr, None))
                     for attr in ('use_cache', 'use_media', 'clear_done',
                                  'target')
                     if getattr(b, attr, None) is not None]

            if attrs:
                msg += ": " + ", ".join(attrs)
            print(msg)

        print_builder(self, level=level - 1)

        if max_level is not None and level > max_level:
            return None

        for subbuilder in self.subbuilders:
            if isinstance(subbuilder, CompositeBuilder):
                subbuilder.print(level=level + 1, max_level=max_level)
            else:
                print_builder(subbuilder, level=level)



"""
The main disseminate program.
"""
import click
import logging

from .options import debug_option
from .init import init
from .build import build
from .preview import preview
from .setup import setup

from ..__version__ import __version__


@click.group()
@click.version_option(__version__, message='v%(version)s')
@debug_option
def main(debug):
    """Disseminate Document Processor"""
    if debug:
        logging.basicConfig(format='%(levelname)-9s:  %(message)s',
                            level=logging.DEBUG)


main.add_command(init)
main.add_command(build)
main.add_command(preview)
main.add_command(setup)



"""
A builder to render a tex file and convert it to an svg.
"""
from .composite_builders import SequentialBuilder


class SvgRender(SequentialBuilder):
    """Render a tex file and render the svg."""

    available = True
    priority = 1000

    infilepath_ext = '.render'
    outfilepath_ext = '.svg'

    def __init__(self, env, parameters=None, outfilepath=None, context=None,
                 template=None, subbuilders=None, use_cache=None, **kwargs):

        # Setup the arguments
        subbuilders = subbuilders or []

        # Setup a PdfRender if no infilepath is specified
        builder_cls = self.find_builder_cls(in_ext='.render', out_ext='.pdf')
        pdfrender = builder_cls(env, context=context, template=template,
                                parameters=parameters, use_cache=True,
                                **kwargs)
        subbuilders.append(pdfrender)
        self.subbuilder_for_outfilename = pdfrender

        # Use the parameter from the pdfrender.
        parameters = parameters or pdfrender.parameters

        # Setup a pdf->svg converter
        builder_cls = self.find_builder_cls(in_ext='.pdf', out_ext='.svg')
        pdf2svg = builder_cls(env, parameters=parameters, use_cache=True,
                              **kwargs)
        subbuilders.append(pdf2svg)

        super().__init__(env, parameters=parameters, outfilepath=outfilepath,
                         subbuilders=subbuilders, use_cache=use_cache,
                         **kwargs)



"""
Software checkers for external dependencies.

Checkers are used to verify the installation and configuration of external
software dependencies.
"""
from .types import (All, Any, Optional,
                    SoftwareDependency, SoftwareDependencyList)
from .checker import Checker
from .python import PythonChecker
from .external import ImageExtChecker
from .pdf import PdfChecker
from .exceptions import MissingHandler

__all__ = ('All', 'Any', 'Optional', 'SoftwareDependency',
           'SoftwareDependencyList', 'Checker', 'PythonChecker',
           'ImageExtChecker', 'PdfChecker', 'MissingHandler')



"""
Signals for label events
"""
from .label_manager import LabelManager
from ..signals import signal


document_onload = signal('document_onload')


@document_onload.connect_via(order=1050)
def reset_label_manager(context, **kwargs):
    """Reset the label manager in the context on document load.

    This should be done before the text of the document is loaded in the
    document object.
    """
    # If no label manager is loaded, then this is the root context. Create a
    # label manager
    label_manager = context.setdefault('label_manager',
                                       LabelManager(root_context=context))

    # Remove labels for this document, so that new labels can be populated
    doc_id = context.get('doc_id', None)
    if doc_id is not None:
        label_manager.reset(doc_ids=doc_id)



"""
A listing of the external dependencies.
"""
from .checkers.types import All, Any, Optional


python_deps = All('python',
                  Any('executables',
                      'python3.6', 'python3.7', 'python3.8', 'python3.9'),
                  All('packages',
                      'regex>=2018.11.22', 'jinja2>=2.11', 'lxml>=4.3.0',
                      'python-slugify>=2.0.1', 'pdfCropMargins>=0.1.4',
                      'click>=7.0', 'tornado>=6.1', 'pygments >=2.6',
                      'pandas>=1.2', 'diskcache>=4.1', 'pathvalidate>=2.2'))

image_deps = Optional('image external dependencies',
                      Optional('executables',
                               'asy', 'convert', 'pdf2svg', 'pdf-crop-margins',
                               'rsvg-convert'))

pdf_deps = All('pdf',
               All('executables',
                   Any('compilers',
                       'pdflatex', 'xelatex', 'lualatex'),
                   Any('package_managers',
                       'kpsewhich',)),
               All('packages',
                   'graphicx', 'caption', 'amsmath', 'mathtools', 'bm',
                   'easylist', 'fancyvrb',
                   'hyperref', 'enumitem', 'geometry', 'xcolor'),
               Optional('fonts',
                        'ecrm1200', 'tcrm1200'),
               Optional('classes',
                        'article', 'report', 'tufte-book', ),
               )



"""
A request handler for rendering the dependency checkers.
"""
from .server import ServerHandler
from ...checkers import Checker, All, Any, Optional


class CheckerHandler(ServerHandler):
    """A request handler for the dependency checkers"""

    def get(self):
        # Get the checkers
        checker_subclses = Checker.checker_subclasses()
        checkers = [checker_subcls() for checker_subcls in checker_subclses]

        # run the checks
        [checker.check() for checker in checkers]

        # Convert the checkers to a list of dicts
        checkers = map(self.checker_to_dict, checkers)

        self.render("checkers.html", checkers=checkers)

    def checker_to_dict(self, item, required=True, level=1):
        """Convert checkers into a list of dicts suitable for rendering the
        view."""
        d = dict()

        d['level'] = level

        # Setup the status string
        available = getattr(item, 'available', None)
        if available is True:
            d['status'] = 'ok'
        elif available is False and required:
            d['status'] = 'fail'
        elif available is False and not required:
            d['status'] = 'missing'
        else:
            d['status'] = 'unknown'

        if isinstance(item, All):
            # All are required
            msg = "Checking required dependencies for '{}'".format(
                item.category)
            d['msg'] = msg
            d['name'] = item.category
            d['subcheckers'] = [
                self.checker_to_dict(i, required=True, level=level + 1)
                for i in item]

        elif isinstance(item, Any):
            # One is required are required
            msg = "Checking alternative dependencies for '{}'".format(
                item.category)
            d['msg'] = msg
            d['name'] = item.category
            d['subcheckers'] = [
                self.checker_to_dict(i, required=False, level=level + 1)
                for i in item]

        elif isinstance(item, Optional):
            # Optional
            # One is required are required
            msg = "Checking alternative dependencies for '{}'".format(
                item.category)
            d['msg'] = msg
            d['name'] = item.category
            d['subcheckers'] = [
                self.checker_to_dict(i, required=False, level=level + 1)
                for i in item]

        elif hasattr(item, 'name'):
            msg = "Checking dependency '{}'".format(item.name)
            d['name'] = item.name
            d['msg'] = msg

        return d



"""
Contexts are dictionaries (dicts) that contain the entries of the header for
a disseminate document, the body of the disseminate document ('body' entry) and
other default settings. The context is used directly in rendering documents.

The context has the following responsibilities:

1. *Variable storage*. The context is a dict that contains the values needed
   to properly render a target document.
2. *Data validation*. Specific entries, like the src_filepath, must be a
   specific type, like a :obj:`SourcePath <disseminate.paths.SourcePath>`.
   When initializing and resetting a context, the context will check that
   entries match this type, and it will convert to the valid type if it can.
3. *Inheritance*. A document and its sub-documents will each have their own
   context. It is useful to have the context of sub-documents inherit from
   the parent document's context (parent context) for some values. A context
   keeps a reference to the parent context, and when resetting the context,
   some (or all) of the parent context values are either copied over or
   referenced.
4. *Intelligent resetting*. A context intelligently resets to its initial
   values and the values of the parent_context on reset. These values can
   then be overwritten when the document source is read in.
"""

from .context import BaseContext

__all__ = ('BaseContext',)



"""
Tags for headings.
"""
from .tag import Tag
from .label import LabelMixin
from .utils import content_to_str
from ..formats import tex_cmd


toc_levels = ('title', 'part', 'chapter', 'section', 'subsection',
              'subsubsection')


class Heading(Tag, LabelMixin):
    """A heading tag.

    .. note::
        If the content isn't specified and an entry exists in the context with
        the tag's name, then this tag's content will be replaced with the
        contents from the context.
    """
    html_name = None
    tex_cmd = None

    active = False
    include_paragraphs = False

    id_mappings = {'title': 'title',
                   'part': 'part',
                   'chapter': 'ch',
                   'section': 'sec',
                   'subsection': 'subsec',
                   'subsubsection': 'subsubsec',
                   }

    def __init__(self, name, content, attributes, context, **kwargs):

        # If no content is specified, see if it's specified in the context
        if (isinstance(content, str) and content.strip() == '' and
           name in context):
            content = content_to_str(context[name])

        # Call the parent class's constructor
        Tag.__init__(self, name=name, content=content, attributes=attributes,
                     context=context, **kwargs)

        if 'nolabel' not in self.attributes:
            LabelMixin.__init__(self, name=name, content=content,
                                attributes=attributes, context=context,
                                **kwargs)

    def generate_label_id(self):
        if 'id' in self.attributes:
            return self.attributes['id']
        else:
            # If an id wasn't specified, automatically generate one and
            # prepend a heading id mapping. 'my-title' to 'ch:my-title'
            cls_name = self.__class__.__name__.lower()
            label_id = LabelMixin.generate_label_id(self)
            return ":".join((Heading.id_mappings[cls_name], label_id))

    def generate_label_kind(self):
        cls_name = self.__class__.__name__.lower()
        return 'heading', cls_name  # ex: ('heading', 'chapter')

    def default_fmt(self, content=None, attributes=None):
        # Prepare the content with the label. References for the default format
        # are not supported
        label_tag = self.label_tag
        if content is not None:
            pass
        elif label_tag is not None:
            content = self.label_tag.default_fmt()
        else:
            content = content_to_str(self.content)

        return super().default_fmt(content=content, attributes=attributes)

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1,
                **kwargs):
        cls_name = self.__class__.__name__.lower()
        content = ''
        attributes = attributes or self.attributes

        # Prepare the tag contents to include the label tag.
        # ex: 'My Title' becomes 'Chap 1. My Title'
        if self.label_tag is not None:
            content += self.label_tag.tex_fmt(mathmode=mathmode,
                                              level=level + 1)

        # Format the heading tag. ex: \chapter{Chapter 1. My First Chapter}
        content = tex_cmd(cls_name, attributes=attributes,
                          formatted_content=content)

        # Get the label for this heading to get its number, if available
        if self.label_id is not None and self.label_anchor is not None:
            assert self.context.is_valid('label_manager')

            label_manager = self.context['label_manager']
            label = label_manager.get_label(id=self.label_id)

            number = label.order[-1] if label.order else None
            number -= 1  # Offset by -1 as the counter is increased
            attrs = ' '.join((cls_name, str(number)))

            content = (tex_cmd('setcounter', attributes=attrs) + '\n' +
                       content + ' ' +
                       self.label_anchor.tex_fmt(mathmode=mathmode,
                                                 level=level + 1))

        return content

    def html_fmt(self, content=None, **kwargs):
        # Prepare the tag contents to include the label tag.
        # ex: 'My Title' becomes 'Chap 1. My Title'
        label_tag = self.label_tag
        if content is not None:
            pass
        elif label_tag is not None:
            content = label_tag
        else:
            content = self.content

        return super().html_fmt(content=content, **kwargs)


class Title(Heading):
    aliases = ("h1",)
    html_name = "h1"
    tex_cmd = "title"
    active = True
    include_paragraphs = False


class Part(Heading):
    html_name = "h1"
    tex_cmd = "part"
    active = True
    include_paragraphs = False


class Chapter(Heading):
    aliases = ("h2",)
    html_name = "h2"
    tex_cmd = "chapter"
    active = True
    include_paragraphs = False


class Section(Heading):
    """A section heading tag."""
    aliases = ("h3", )
    html_name = "h3"
    tex_cmd = "section"
    active = True
    include_paragraphs = False


class SubSection(Heading):
    """A subsection heading tag."""
    aliases = ("h4",)
    html_name = "h4"
    tex_cmd = "subsection"
    active = True
    include_paragraphs = False


class SubSubSection(Heading):
    """A subsubsection heading tag."""
    aliases = ("h5",)
    html_name = "h5"
    tex_cmd = "subsubsection"
    active = True
    include_paragraphs = False


class Para(Tag):
    """A paragraph heading tag."""
    aliases = ("h6",)

    html_name = "paragraph-heading"
    tex_cmd = "paragraph"
    active = True
    include_paragraphs = False

    label_heading = False



"""
Tags for figure captions and references.
"""
from .tag import Tag
from .label import LabelMixin
from .utils import content_to_str
from ..formats import tex_cmd
from ..utils.string import hashtxt


class CaptionError(Exception):
    """An error was encountered in processing a caption."""
    pass


class Caption(Tag, LabelMixin):
    """A tag for captions.

    .. note:: The use of a naked caption tag is allowed (i.e. a caption not
              nested within a figure or table), but this won't register a label
              with the label manager. In order to create a label, the
              create_label method in the LabelMixin must be invoked.

              :example:

                 ::

                    @fig{@img{image.svg}
                         @caption{My first figure}
                        }
    """

    # Render captions as <span> because the <caption> tag is a block element,
    # and captions can be rendered inline
    html_name = 'span'
    html_class = 'caption'
    tex_cmd = 'caption'

    kind = None
    active = True

    def __init__(self, *args, **kwargs):
        # Call the tag constructor, but not the LabelMixin construction.
        # The create_label method of LabelMixin must be invoked separately.
        Tag.__init__(self, *args, **kwargs)

        # Set the attributes to the class
        if 'class' not in self.attributes:
            self.attributes['class'] = 'caption'

    def generate_label_id(self):
        """Generate the label id and set the id in the attributes."""
        if self.label_id is not None:
            label_id = self.label_id
        elif 'id' in self.attributes:
            label_id = self.attributes['id']
        else:
            text = self.default_fmt()
            label_id = 'caption-' + hashtxt(text)

        # Set the 'id' attributes
        self.attributes['id'] = label_id
        return label_id

    def generate_label_kind(self):
        return self.kind if self.kind is not None else ('caption',)

    def default_fmt(self, attributes=None, content=None):
        # Prepare the content with the label. References for the default format
        # are not supported
        content = ''
        if self.label_tag is not None:
            content += self.label_tag.default_fmt()
        content += content_to_str(self.content)

        return super().default_fmt(content=content, attributes=attributes)

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1):
        cls_name = self.__class__.__name__.lower()
        content = ''

        # Prepare the tag contents to include the label tag.
        # ex: 'My Title' becomes 'Chap 1. My Title'
        if self.label_tag is not None:
            content += self.label_tag.tex_fmt(mathmode=mathmode,
                                              level=level)

        # Format the tag's contents into a string
        content += content_to_str(self.content, target='.tex',
                                  mathmode=mathmode, level=level)

        # Format the heading tag. ex: \chapter{Chapter 1. My First Chapter}
        content = tex_cmd(cls_name, attributes=self.attributes,
                          formatted_content=content)

        if self.label_anchor is not None:
            content += ' ' + self.label_anchor.tex_fmt(mathmode=mathmode,
                                                       level=level)

        return content

    def html_fmt(self, content=None, **kwargs):
        # Prepare the tag contents to include the label tag.
        # ex: 'My Title' becomes 'Chap 1. My Title'
        content = []
        if self.label_tag is not None:
            content.append(self.label_tag)

        if isinstance(self.content, list):
            content += self.content
        else:
            content.append(self.content)

        return super().html_fmt(content=content, **kwargs)



"""
Receivers for processing a tag's paragraphs on tag creation.
"""
import regex

from ..signals import tag_created


@tag_created.connect_via(order=400)
def process_paragraphs(tag, tag_factory, tag_base_cls, **kwargs):
    """A receiver to parse the paragraphs of tags."""
    if tag.name in tag.context.get('process_paragraphs', []):
        context = tag.context
        p_cls = tag_factory.tag_class(tag_name='P', context=context)
        process_paragraph_tags(tag, context=tag.context,
                               tag_base_cls=tag_base_cls,
                               p_cls=p_cls)
    return tag


re_para = regex.compile(r'(?:\s*\n\s*\n\s*\n*)')


def group_paragraphs(elements):
    r"""Given a list, group the items into sublists based on strings with
    newlines.

    .. note:: This function will not include items in the ast, including tags,
              in paragraphs sublists that have an attribute
              'include_paragraphs' with a value of False.

    .. note:: The function is idempotent. It will reprocess the generated AST
              and not make changes.

    Parameters
    ----------
    elements : Union[list, str]
        A string or a list of items and strings to group paragraphs.

    Returns
    -------
    parse_list : list
        The list with sublists denoting paragraph breaks.

    Examples
    --------
    >>> group_paragraphs([1, 2, 'three', 'four\n\nfive', 6,
    ...                   'seven\n\neight'])
    [[1, 2, 'three', 'four'], ['five', 6, 'seven'], ['eight']]
    >>> group_paragraphs('This is my\n\ntest paragraph.')
    [['This is my'], ['test paragraph.']]
    """
    # Wrap strings in a list
    elements = [elements] if isinstance(elements, str) else elements

    overall_list = []
    sublist = []

    for item in elements:

        # Add non-string elements to the sublist
        if not isinstance(item, str):
            if (getattr(item, 'include_paragraphs', True) and
               not isinstance(item, list)):
                # Include in the paragraph sublist
                sublist.append(item)
            else:
                # Do not include in paragraphs; create a new paragraph sublist
                if sublist:
                    overall_list.append(sublist)
                    sublist = []
                # Append this non-paragraph item to the overall list, outside
                # of a paragraph sublist
                overall_list.append(item)
            continue

        # At this point, item is a string. See if there a paragraph break in
        # the string.
        pieces = re_para.split(item)

        if len(pieces) == 1 and pieces[0]:
            # In this case, no paragraph break was found. Just add the item
            # to the sublist if it's not an empty string
            sublist.append(pieces[0])
            continue

        # In this case, multiple pieces were found. Make these into new
        # sublists
        for i, piece in enumerate(pieces):
            if piece:
                # Add the piece to the paragraph sublist, if it's not an
                # empty string
                sublist.append(piece)

            if i != len(pieces) - 1:
                if sublist:
                    overall_list.append(sublist)
                sublist = []

    if sublist:
        overall_list.append(sublist)

    elements.clear()
    elements += overall_list

    return elements


def clean_paragraphs(elements):
    r"""Remove invalid paragraphs from the sublists in an ast created by
    group_paragraphs.

    This function will:

    1. Remove sublists created by group_paragraphs that only contain empty
       strings and strings with space or newline characters.

    Parameters
    ----------
    elements : Union[list, str]
        A string or a list of items and strings to group paragraphs.

    Returns
    -------
    cleaned_list : list
        The list with sublists denoting paragraph breaks.

    Examples
    --------
    >>> group = group_paragraphs([1, 2, 'three', 'four\n\nfive', 6,
    ...                           'seven\n\neight'])
    >>> clean_paragraphs(group)
    [[1, 2, 'three', 'four'], ['five', 6, 'seven'], ['eight']]
    >>> group = group_paragraphs([1, 2, 'three', '\n\n', 6,
    ...                           'seven\n\neight'])
    >>> clean_paragraphs(group)
    [[1, 2, 'three'], [6, 'seven'], ['eight']]
    """
    assert isinstance(elements, list)

    new_elements = []

    for item in elements:
        # Determine if item is a sublist with only empty strings or strings
        # with space and newline characters. If so, don't make a paragraph
        # with it, and skip it.
        if (isinstance(item, list) and
           all(isinstance(i, str) and not i.strip() for i in item)):
            continue

        new_elements.append(item)

    # Copy over the new_ast to the given ast
    elements.clear()
    elements += new_elements

    return elements


def assign_paragraph_roles(elements, tag_base_cls):
    """Assign the 'paragraph_role' attribute for tags within sublists created
    by the group_paragraphs function.

    The 'paragraph_role' attribute can either be 'block', for tags that are
    alone in a paragraph, or 'inline' for tags that are in a paragraph with
    other text elements.

    Parameters
    ----------
    elements : Union[str, list, :obj:`Tag <.Tag>`]
        The list of paragraph tags (:obj:`P <.text.P>` and
        strings.
    tag_base_cls : :class:`Tag <.Tag>`
        The base class for Tag objects.

    Returns
    -------
    processed_ast : list
        The list with tags in the sublists marked with their paragraph_role.
    """
    assert isinstance(elements, list)

    # Go over the paragraph sublists and determine whether the tags within
    # are inline or block
    for sublist in filter(lambda x: isinstance(x, list), elements):
        # Find all the tags in the sublist
        sublist_tags = list(filter(lambda x: isinstance(x, tag_base_cls),
                                   sublist))

        # Find all strings without white space
        sublist_string = [s for s in sublist
                          if isinstance(s, str) and s.strip() != ""]

        # Determine the number of tags and the number of total elements in
        # the sublist
        num_tags = len(sublist_tags)
        num_elems = len(sublist)
        num_nonempty_strings = len(sublist_string)

        if num_tags == 1 and num_nonempty_strings == 0:
            # If the number of tags and elements is 1, then tag is in its own
            # block
            sublist_tags[0].paragraph_role = 'block'
        elif num_elems > 1:
            # Otherwise the tags are inline with other elements in the
            # paragraph.
            for tag in sublist_tags:
                tag.paragraph_role = 'inline'

    return elements


def process_paragraph_tags(element, context, tag_base_cls, p_cls):
    """Process the paragraphs for the contents of a tag.

    Parameters
    ----------
    element : Union[str, list, :obj:`Tag <.Tag>`]
        A string, tag or list of both to process for paragraphs
    context : :obj:`DocumentContext <.DocumentContext>`
        The context with values for the document.
    tag_base_cls : :class:`Tag <.Tag>`
        The base class for Tag objects.
    p_cls : :class:`P <.text.P>`
        The tag class for paragraphs.

    Returns
    -------
    processed_contents : Union[str, :obj:`Tag <.Tag>`, list]
        The contents with
    """
    if isinstance(element, tag_base_cls):
        # If it's a tag, pull out its content
        content = element.content
    else:
        # Otherwise use the element directly as the content to process
        content = element

    # Check the format of the content. This function can only process strings
    # or lists
    if not any(isinstance(content, x) for x in (str, list)):
        return content

    # Group the paragraphs into sublists
    group = group_paragraphs(content)

    # Clean the paragraph sublist groups
    clean_paragraphs(group)

    # Assign the paragraph_role for tags within paragraph sublist groups
    assign_paragraph_roles(group, tag_base_cls=tag_base_cls)

    # Convert the sublists into paragraphs
    for count, item in enumerate(group):
        if isinstance(item, list):
            # If the item is a list with only 1 item, then isolate that one
            # item. The paragraph will then only contain that one item, rather
            # than have a list with one item.
            item = item[0] if len(item) == 1 else item

            # sublists created by group_paragraphs are paragraphs
            p = p_cls(name='p', content=item, attributes='', context=context)

            group[count] = p

    # wrap the group, if it's only 1 item
    if len(group) == 1:
        group = group[0]

    # If the original element was a tag, replace its content
    if isinstance(element, tag_base_cls):
        element.content = group
        return element
    else:
        return group



"""
Code formatting tags
"""
from pygments import highlight, lexers, formatters
from markupsafe import Markup

from .tag import Tag
from .utils import content_to_str
from ..utils.types import StringPositionalValue
from ..paths.utils import find_files
from ..formats import xhtml_tag, tex_verb


class Code(Tag):
    """A tag for displaying code.
    """

    active = True

    process_content = False  # Do not process the contents
    process_typography = False  # Do not process typography tags
    include_paragraphs = True  # Do not include paragraphs within

    html_name = 'code'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load files, if needed
        files = find_files(self.content, self.context)

        if len(files) > 0:
            new_content = "\n".join([file.read_text() for file in files])
            self.content = new_content

    @property
    def lexer(self):
        """Get the lexer for code highlighting."""
        # Get string positional values from the attributes for the kind of
        # code
        string_positionals = [k for k, v in self.attributes.items()
                              if v == StringPositionalValue]

        # See if a lexer is available
        for string_positional in string_positionals:
            # See if a lexer is available
            lexer_name = string_positional.title() + 'Lexer'

            if lexer_name in lexers.LEXERS:
                return lexers.get_lexer_by_name(string_positional)

        # Otherwise try to guess it from the code block contents
        return lexers.guess_lexer(self.content)

    def html_fmt(self, content=None, attributes=None, format_func='html_fmt',
                 method='html', level=1):
        # Format contents
        content = self.content if content is None else content
        content = content_to_str(content=content)

        # Get the lexer for the code and formatter for html
        if self.paragraph_role == 'block':
            lexer = self.lexer
            cssclass = ('highlight' if self.paragraph_role is None else
                        'highlight ' + self.paragraph_role)
            formatter = formatters.HtmlFormatter(cssclass=cssclass)

            html = highlight(content, lexer, formatter)
            return xhtml_tag('div', attributes='class="code"',
                             formatted_content=Markup(html), method=method,
                             level=level)
        else:
            return xhtml_tag('code', attributes=self.attributes,
                             formatted_content=content, method=method,
                             level=level)

    def tex_fmt(self, content=None, attributes=None, mathmode=False, level=1):
        # Format contents
        content = self.content if content is None else content
        content = content_to_str(content=content)

        if self.paragraph_role == 'block':
            lexer = self.lexer
            return highlight(content, lexer, formatters.LatexFormatter())
        else:
            return tex_verb(content)


class Python(Code):
    """A tag for displaying python code"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes['python'] = StringPositionalValue


class Html(Code):
    """A tag for displaying html code"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes['html'] = StringPositionalValue


class Ruby(Code):
    """A tag for displaying ruby code"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes['ruby'] = StringPositionalValue


class Java(Code):
    """A tag for displaying ruby code"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes['java'] = StringPositionalValue


class Javascript(Code):
    """A tag for displaying javascript code"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes['javascript'] = StringPositionalValue


class Dm(Code):
    """A tag for displaying dm code"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attributes['dm'] = StringPositionalValue



"""
Signals for target builders
"""
from ...signals import signal

add_file = signal("add_file",
                  doc="Add a file dependency to a target builder. Takes "
                      "parameters, context, in_ext, target and use_cache.")

find_builder = signal("find_builder",
                      doc="Given a document context and a target, find the "
                          "corresponding target builder.")



"""
A receiver to set the document label.
"""
from ..signals import document_onload
from ...utils.string import slugify


@document_onload.connect_via(order=1100)
def process_document_label(context, **kwargs):
    """A context processor to set the document label in the label manager."""
    assert context.is_valid('label_manager', 'document', 'src_filepath')

    label_manager = context['label_manager']
    src_filepath = context['src_filepath']
    doc = context['document']()  # de-reference weakref

    subpath_str = str(src_filepath.subpath)

    # Get the level of the document
    level = context.get('level', 1)

    # Generate the label_id. ex: 'sub/file1.dm' becomes 'doc:sub-file1-dm'
    label_id = 'doc:' + slugify(subpath_str)

    # Set the label for this document
    kind = ('document', 'document-level-' + str(level))
    label_manager.add_document_label(id=label_id,
                                     kind=kind, title=doc.short,
                                     context=context)



"""
Utility functions for dictionaries.
"""


def find_entry(dicts, *keys, suffix=None, sep='_'):
    """Search dictionaries for entries with string keys constructed from
    ordred combinations of the specified keys.

    Dictionaries are searched in order of preference.
    String keys are constructed with parts constructed by joining the parts
    with the sep string.

    Parameters
    ----------
    dicts : Union[dict, Tuple[dict], List[dict]]
        One or more dictionaries to find entries in.
    keys : str
        One or more items to construct keys.
    suffix : Optional[str]
        If specified, try appending the given suffix to generate the combined
        key before trying the combined key itself.
    sep : Optional[str]
        The separator for joining key parts into a key.

    Returns
    -------
    value
        The value from the dict for the best match.

    Raises
    ------
    KeyError
        Raised if a key was not found.

    Examples
    --------
    >>> d = {'starter_middle1_term': 1,
    ...      'starter_middle1': 2,
    ...      'starter_middle2_term': 3,
    ...      'starter_middle2': 4,
    ...      'starter_term': 5,
    ...      'starter': 6}
    >>> find_entry(d, 'starter', 'middle1', suffix='term')
    1
    >>> find_entry(d, 'starter', 'middle1')
    2
    >>> find_entry(d, 'starter', 'middle2', suffix='term')
    3
    >>> find_entry(d, 'starter', 'middle2' )
    4
    >>> find_entry(d, 'starter', 'term')
    5
    >>> find_entry(d, 'starter')
    6
    """
    # Wrap dict in a list
    dicts = [dicts] if isinstance(dicts, dict) else dicts

    for d in dicts:
        num_keys = len(keys)
        while num_keys > 0:
            trial = sep.join(keys[0:num_keys])
            trial_suffix = (sep.join((trial, suffix))
                            if suffix is not None else trial)

            for t in (trial_suffix, trial):
                if t in d:
                    return d[t]

            num_keys -= 1

    msg = "Could not find a key constructed from '{keys}'"
    raise KeyError(msg.format(keys=keys))



"""
Tags are interpreted text elements in a disseminate document that are converted
to other elements in the rendered document, like bold text, images, figures,
equations and tables.
"""

from .tag import Tag, TagFactory
from .exceptions import TagError
from . import signals, receivers
from . import (headings, text, img, asy, notes, figs, caption, code, eqs,
               featurebox, toc, list, preamble, label, ref, navigation, data,
               table)

__all__ = ('Tag', 'TagFactory', 'TagError', 'signals', 'receivers',
           'headings', 'text', 'img', 'asy', 'notes', 'figs', 'caption',
           'code', 'eqs', 'featurebox', 'toc', 'list', 'preamble', 'label',
           'ref', 'navigation', 'data', 'table')



"""
A receiver to process headers in a context
"""
import regex
import pathlib

from ..signals import document_onload
from ...context import BaseContext
from ...context.utils import find_header_entries, load_from_string
from ...utils.list import uniq
from ... import settings


@document_onload.connect_via(order=1000)
def process_headers(context, **kwargs):
    """Process header strings for entries in a context by loading them into
    the context.

    Context information comes from the following sources:

    1. The document itself and its header file
    2. Additional headers from the template

    The information in (1) takes precedence over (2). For this reason, (1)
    should be loaded after (2), but (2) is specified by some entries, like
    renderers, template, targets, in (1), so (1) is pre-loaded before (2).
    """
    assert context.is_valid('paths')

    # 1. Load the context entries with headers
    # See which context entries have a header
    keys = find_header_entries(context)

    # Load the context from the document's header
    header_context = BaseContext()
    for key in keys:
        rest, d = load_from_string(context[key])

        # Replace the entry with the context removed
        context[key] = rest

        # Update the context
        header_context.match_update(d)

    # 2. Load the template paths and the 'context.txt' files
    #    First load the template paths for the template and any parent
    #    templates. The template_name may be different from the header_context
    #    than the one specified in this context.
    template_name = header_context.get('template') or context.get('template')
    template_paths = find_template_paths(template_name=template_name)

    for template_path in list(template_paths):
        template_paths += find_jinja2_parent_templates(template_path)
    template_paths = uniq(template_paths)

    # Next, find the additional context files and load their context values.
    # These are done in *reverse* order because the parent templates are
    # listed last and the child templates listed first. The child template
    # values take precedence.
    template_context = BaseContext()

    # Get the additional context filepaths in reverse order
    fps = find_additional_context_filepaths(template_paths[::-1])
    for context_filepath in fps:
        template_context.load(context_filepath.read_text())

    # Now load the template paths and context values
    paths = context['paths']
    paths[0:0] = template_paths  # insert at top of list w/o creating new list

    # Copy over entries in the template_context, first, then those from the
    # header_context. Since the header_context contains entries from the
    # headers of source files, these take precedence over template values
    context.match_update(template_context, overwrite=True)
    context.match_update(header_context, overwrite=True)


def find_template_paths(template_name):
    """Find template paths from a template name.

    Parameters
    ----------
    template_name : str
        The name of the template. ex: books/tufte

    Returns
    -------
    template_paths : List[:obj:`pathlib.Path`]
        A list of template path directories for the template.
    """
    paths = []

    # find template paths
    for template_basename in settings.module_template_paths:
        template_path = template_basename / template_name

        if template_path.is_dir() and template_path not in paths:
            # for template='books/tufte', for example
            # Add template plate to the front
            paths.insert(0, template_path)
        elif (template_path.parent.is_dir() and
              template_path.parent not in paths):
            # for template='default/template', for example
            paths.insert(0, template_path.parent)

    return uniq(paths)


_re_jinja2_extends = regex.compile(r'extends\s*[\"\']([^\'\"]+)[\"\']')


def find_jinja2_parent_templates(template_path):
    """Find additional templates from jinja2 template paths.

    Parameters
    ----------
    template_path : :obj:`pathlib.Path`
        The path for the directory of the child template.

    Returns
    -------
    template_paths : List[:obj:`pathlib.Path`]
        A list of template path directories for the parent templates.
    """
    # Find all template files
    template_files = template_path.glob('**/template*')

    # Find "extends" statements. This regex will find items
    # like {% extends "default/template.html" %}
    matches = [_re_jinja2_extends.search(tf.read_text())
               for tf in template_files]

    # Remove None enties
    matches = filter(bool, matches)  # remove None entries

    # Find the paths
    paths = []
    for match in matches:
        # Format the template name: Strip the template filename and
        # target directory
        # ex: 'default/xhtml/template.xhtml' -> 'default'
        template_name = match.group(1)
        template_name = pathlib.Path(template_name).parent.parent

        # Add the path as long as the above didn't return the current
        # directory, which is not a template
        if str(template_name) != '.':
            paths += find_template_paths(template_name)

    return uniq(paths)


def find_additional_context_filepaths(template_paths, context_filename=None):
    """Find additional context files (context.txt) for the given template path.

    Parameters
    ----------
    template_paths : List[:obj:`pathlib.Path`]
        The list of template directory paths to search for additional context
        files
    context_filename : Optional[str]
        The default filename for the additional context files in the template
        directories (ex: context.txt)

    Returns
    -------
    additional_context_paths : List[:obj:`pathlib.Path`]
        A list of filepaths for additional context files in the template
        directories.
    """
    context_filename = context_filename or settings.template_context_filename
    additional_context_filepaths = []

    for template_path in template_paths:
        context_filepath = template_path / context_filename
        if context_filepath.is_file():
            additional_context_filepaths.append(context_filepath)
            continue

        context_filepath = template_path.parent / context_filename
        if context_filepath.is_file():
            additional_context_filepaths.append(context_filepath)
            continue

    return uniq(additional_context_filepaths)



"""
The Ref tag to reference captions and other labels.
"""
from .tag import Tag
from .exceptions import assert_content_str
from .utils import content_to_str, format_content
from ..label_manager.types import DocumentLabel
from ..signals import signal
from ..formats import xhtml_tag, tex_cmd


ref_label_dependencies = signal("ref_label_dependencies")


@ref_label_dependencies.connect_via(order=1000)
def add_ref_labels(builder, **kwargs):
    """Find and add the labels associated with Ref tags for all tags in
    the context.
    """
    context = builder.context

    # Find tags in the context and get the ref tag labels
    ref_label_ids = set()
    for tag in filter(lambda t: isinstance(t, Tag), context.values()):
        # Flatten the tag tree
        flat_tags = tag.flatten(filter_tags=True)

        # Retrieve all Ref tags and their label ids
        ids = [t.label_id for t in flat_tags if isinstance(t, Ref)]
        ref_label_ids.update(ids)

    # Remove None entries
    ref_label_ids.discard(None)

    # Retrieve the labels
    label_man = context.get('label_manager')
    if label_man is not None and ref_label_ids:
        return sorted(label_man.get_labels_by_id(ids=ref_label_ids))
    else:
        return []


class RefError(Exception):
    """A reference to a document could not be found."""
    pass


class Ref(Tag):
    """A tag to reference a label.

    Attributes
    ----------
    doc_id : str
        The doc_id for the document that owns the label referenced. This
        might be different to the doc_id listed in the context if a reference
        is made to a label in another document.
    label_id : str
        The id of the label referenced by this tag.

    .. note:: The ref tag takes the label id as the content. If the label is
              not found, a label manager exception is raised.
              (:exc:`.label_manager.exceptions.LabelNotFound`)
    """

    active = True

    doc_id = None
    label_id = None

    def __init__(self, name, content, *args, **kwargs):
        assert_content_str(content)

        super(Ref, self).__init__(name, content, *args, **kwargs)

        # Set the label_id and doc_id
        self.label_id = self.content.strip()

    @property
    def label(self):
        """Retrieve the label for this ref tag."""
        assert self.context.is_valid('label_manager')
        label_manager = self.context['label_manager']
        if self.label_id is not None:
            return label_manager.get_label(id=self.label_id)
        else:
            return None

    def document(self, cache=None):
        """The document that owns the label referenced by this tag.

        Parameters
        ----------
        cache : Optional[dict]
            If specified, the cache values will be used instead of being
            evaluated. Possibilities:
            - 'label': :obj:`.types.Label`
            - 'documents_by_id': Dict[str, :obj:`document.Document`]

        Returns
        -------
        document : Union[:obj:`Document <.Document>`, None]
            The document that owns the label referenced by this tag, if
            available.
            None, if a document could not be found.
        """
        assert self.context.is_valid('root_document')
        cache = dict() if cache is None else cache
        label = cache.setdefault('label', self.label)
        documents_by_id = cache.get('documents_by_id', None)

        # Get the doc_ids for the document that owns this tag (doc_id) and
        # the document that owns the label (other_doc_id)
        doc_id = self.context.get('doc_id', None)
        other_doc_id = label.doc_id

        if doc_id is not None and doc_id == other_doc_id:
            # Return this document, if the label is owned by the same document
            # as the owner of this tag.
            return self.context.document

        # The other_doc_id and doc_id are different, so this Ref tag references
        # a label owned by another document.
        #
        # Fetch the root document to figure out which document corresponds to
        # the doc_id of this ref tag's label.
        if documents_by_id is None:
            root_document = (self.context['root_document']()
                             if 'root_document' in self.context else
                             None)  # de-reference root document, if available
            docs_by_doc_ids = (root_document.documents_by_id(recursive=True)
                               if root_document is not None else None)
        else:
            docs_by_doc_ids = documents_by_id

        if other_doc_id in docs_by_doc_ids:
            return docs_by_doc_ids[other_doc_id]
        else:
            return None

    def url(self, target='.html', include_anchor=True, cache=None):
        """The url path for the document referenced by the label for this tag.

        Parameters
        ----------
        target : Optional[str]
            The target extension for the target file.
            ex: '.html' or '.tex'
        include_anchor : Optional[bool]
            If True (default), the html link anchor will be appended to the
            url path.
        cache : Optional[dict]
            If specified, the cache values will be used instead of being
            evaluated. Possibilities:
            - 'label': :obj:`.types.Label`
            - 'documents_by_id': Dict[str, :obj:`document.Document`]

        Returns
        -------
        url_path : str
            The url path for the document owning the label referenced by this
            tag.
        """
        context = self.context
        cache = dict() if cache is None else cache
        assert context.is_valid('doc_id')

        # Format the target string
        target = (target if target.startswith('.') else '.' + target)

        # Get the doc_ids for the document that owns this tag (doc_id) and
        # the document that owns the label (other_doc_id). See if they're the
        # same document, in which case an internal link is returned.
        label = cache.setdefault('label', self.label)
        doc_id = self.context.get('doc_id', None)
        other_doc_id = label.doc_id if label is not None else None

        if doc_id is None or other_doc_id is None:
            msg = ("Could not find the tag or label document for the ref tag "
                   "'{}'")
            raise RefError(msg.format(self))
        elif doc_id is not None and doc_id == other_doc_id:
            return '#' + label.id if include_anchor else ''

        # In this case, the label and tag documents are different. Return a
        # link to the label's document.
        document = self.document(cache=cache)
        target_filepath = document.target_filepath(target)

        if target_filepath is None:
            return None

        link = target_filepath.get_url(context=self.context)

        return link + '#' + label.id if include_anchor else link

    def default_fmt(self, content=None, attributes=None, cache=None, **kwargs):
        """Convert the tag to a text string.

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.Tag>`]]]
            Specify an alternative content from the tag's content. It can
            either be a string, a tag or a list of strings, tags and lists.
        attributes : Optional[Union[str, :obj:`Attributes <.Attributes>`]]
            Specify an alternative attributes dict from the tag's attributes.
            It can either be a string or an attributes dict.
        cache : Optional[dict]
            If specified, the cache values will be used instead of being
            evaluated. Possibilities:
            - 'label': :obj:`.types.Label`
            - 'documents_by_id': Dict[str, :obj:`document.Document`]

        Returns
        -------
        text_string : str
            A text string with the tags stripped.
        """
        # Get the label tag format
        label_manager = self.context.get('label_manager')
        context = self.context
        cache = dict() if cache is None else cache
        label = cache.setdefault('label', self.label)

        if all(i is not None for i in (label_manager, label, context)):
            # Format the format string keys for a ref
            keys = ('ref', *label.kinds)
            format_str = label_manager.format_string(id=self.label.id, *keys)

            processed_tag = Tag(name='ref', content=format_str, attributes='',
                                context=context)
            return content_to_str(processed_tag.content)
        else:
            return ''

    def tex_fmt(self, content=None, attributes=None, mathmode=False,
                cache=None, level=1, **kwargs):
        """Format the tag in LaTeX format.

        .. note:: This function renders tex links to compiled pdf documents

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.Tag>`]]]
            Specify an alternative content from the tag's content. It can
            either be a string, a tag or a list of strings, tags and lists.
        attributes : Optional[Union[str, :obj:`Attributes <.Attributes>`]]
            Specify an alternative attributes dict from the tag's attributes.
            It can either be a string or an attributes dict.
        mathmode : Optional[bool]
            If True, the tag will be rendered in math mode. Otherwise (default)
            latex text mode is assumed.
        cache : Optional[dict]
            If specified, the cache values will be used instead of being
            evaluated. Possibilities:
            - 'label': :obj:`.types.Label`
            - 'documents_by_id': Dict[str, :obj:`document.Document`]
        level : Optional[int]
            The level of the tag.

        Returns
        -------
        tex_string : str
            The formatted tex string.
        """
        label_manager = self.context.get('label_manager')
        context = self.context
        cache = dict() if cache is None else cache
        label = cache.setdefault('label', self.label)

        if all(i is not None for i in (label_manager, label, context)):
            # Retrieve the format string for the reference
            keys = ('ref', *label.kind)
            format_str = label_manager.format_string(label.id, *keys,
                                                     target='.tex')

            # process the tags and format the contents
            # for html
            processed_tag = Tag(name='ref', content=format_str, attributes='',
                                context=context)
            content = format_content(content=processed_tag.content,
                                     format_func='tex_fmt', level=level + 1)

            # setup a url path and include the anchor if the label is not
            # for a DocumentLabel. DocumentLabels should just point to the
            # file itself.
            # Tex formats will only work with pdf links
            include_anchor = not isinstance(label, DocumentLabel)

            url = self.url(target='.pdf', include_anchor=include_anchor,
                           cache=cache)

            # Add a target-specific attribute to the url so that it's
            # properly parsed for the '.tex' target
            url = url + '.tex' if url else ''

            # wrap content in 'href' tag, if a url is present, otherwise
            # just return a regular text string
            return (tex_cmd('href', attributes=url, formatted_content=content)
                    if url else content)
        else:
            return ''

    def html_fmt(self, content=None, attributes=None, cache=None,
                 format_func='html_fmt', method='html', level=1, **kwargs):
        """Convert the tag to an (x)html string or (x)html element.

        Parameters
        ----------
        content : Optional[Union[str, List[Union[str, list, :obj:`Tag \
            <.Tag>`]]]
            Specify an alternative content from the tag's content. It can
            either be a string, a tag or a list of strings, tags and lists.
        attributes : Optional[Union[str, :obj:`Attributes <.Attributes>`]]
            Specify an alternative attributes dict from the tag's attributes.
            It can either be a string or an attributes dict.
        cache : Optional[dict]
            If specified, the cache values will be used instead of being
            evaluated. Possibilities:
            - 'label': :obj:`.types.Label`
            - 'documents_by_id': Dict[str, :obj:`document.Document`]
            - 'format_str': str
        format_func : Optional[str]
            The tag format function to use in rendering the reference tag.
            (ex: 'html_fmt' or 'xhtml_fmt')
        method : Optional[str]
            The rendering method for the string. ex: 'html' or 'xml'
        level : Optional[int]
            The level of the tag.

        Returns
        -------
        html : str or html element
            A string in HTML format or an HTML element (:obj:`lxml.builder.E`).
        """
        label_manager = self.context.get('label_manager')
        cache = dict() if cache is None else cache
        label = cache.setdefault('label', self.label)
        context = self.context

        if all(i is not None for i in (label_manager, label, context)):

            # Retrieve the format string for the reference
            # Use the 'html' format_str for html and xhtml
            keys = ('ref', *label.kind)
            format_str = label_manager.format_string(label.id, *keys,
                                                     target='.html')

            # substitute the link, process the tags and format the contents
            # for html (html_fmt) or xhtml (xhtml_fmt)
            processed_tag = Tag(name='ref', content=format_str, attributes='',
                                context=context)
            content = format_content(content=processed_tag.content,
                                     format_func=format_func, level=level + 1)

            attrs = (self.attributes.copy() if attributes is None else
                     attributes)
            attrs['class'] = 'ref'

            # setup the url path and include the anchor if the label is not
            # for a DocumentLabel. DocumentLabels should just point to the
            # file itself.
            include_anchor = not isinstance(label, DocumentLabel)
            attrs['href'] = self.url(target=method, cache=cache,
                                     include_anchor=include_anchor)

            # wrap content in 'a' tag
            return xhtml_tag('a', attributes=attrs, formatted_content=content,
                             level=level, method=method,
                             pretty_print=False)  # no line breaks
        else:
            return ''



"""
Signals for document events
"""

from ..signals import signal

document_created = signal('document_created',
                          doc="Signal sent after document creation. "
                              "Receivers take a document parameter.")
document_deleted = signal('document_deleted',
                          doc="Signal sent before a document is deleted. "
                          "Receivers take a document parameter.")
document_onload = signal('document_onload',
                         doc="Signal sent when a document is loaded. "
                         "Receivers take a document or document context "
                         "parameter.")

document_build = signal('document_build',
                        doc="Signal sent when a document's targets are "
                        "being built to their final target files. "
                        "Receivers take a document parameter.")

document_build_needed = signal('document_build_needed',
                               doc="Signal sent to evaluate whether a build "
                               "is needed. Takes a document as a parameter "
                               "and returns True or False")

document_tree_updated = signal('document_tree_updated',
                               doc="Signal sent when a root document or one "
                               "of its sub-documents was re-loaded. Takes a "
                               "root document as a parameter.")



"""
A builder to convert from PDF to SVG
"""
from .builder import Builder
from .composite_builders import SequentialBuilder
from .pdfcrop import PdfCrop
from .scalesvg import ScaleSvg


class Pdf2svg(Builder):
    """A builder to convert from pdf to svg."""

    action = "pdf2svg {builder.infilepaths} {builder.outfilepath}"
    available = False
    priority = 1000
    required_execs = ('pdf2svg',)

    infilepath_ext = '.pdf'
    outfilepath_ext = '.svg'

    page_no = None

    def __init__(self, env, **kwargs):
        super().__init__(env, **kwargs)

        # Get the page number
        page_no = self.get_parameter('page') or self.get_parameter('page_no')
        try:
            self.page_no = int(page_no)
        except TypeError:
            pass

    def run_cmd_args(self):
        args = list(super().run_cmd_args())
        if isinstance(self.page_no, int):
            args += [self.page_no]
        return tuple(args)


class Pdf2SvgCropScale(SequentialBuilder):
    """A SequentialBuilder for Pdf2Svg that includes PdfCrop and ScaleSvg
    builders.
    """

    available = True
    priority = 1000
    infilepath_ext = '.pdf'
    outfilepath_ext = '.svg'

    def __init__(self, env, parameters=None, subbuilders=None, use_cache=None,
                 **kwargs):
        # Setup parameters. The parameters list should be a copy.
        subbuilders = (list(subbuilders) if (isinstance(subbuilders, list) or
                       isinstance(subbuilders, tuple)) else [])
        parameters = parameters if parameters is not None else []
        parameters = (list(parameters) if isinstance(parameters, tuple) or
                      isinstance(parameters, list) else [parameters])

        # Create the subbuilders
        crop = self.get_parameter('crop', *parameters)
        crop = (crop if crop is not None else
                self.get_parameter('crop_percentage', *parameters))

        if crop is not None:
            pdfcrop = PdfCrop(env, parameters=parameters, use_cache=True,
                              **kwargs)
            subbuilders.append(pdfcrop)

        pdf2svg = Pdf2svg(env, use_cache=True, **kwargs)
        subbuilders.append(pdf2svg)

        scale = self.get_parameter('scale', *parameters)
        if scale is not None:
            scalesvg = ScaleSvg(env, parameters=parameters, use_cache=True,
                                **kwargs)
            subbuilders.append(scalesvg)

        super().__init__(env, parameters=parameters, subbuilders=subbuilders,
                         use_cache=use_cache, **kwargs)



"""
A builder to scale svg
"""
from .builder import Builder


class ScaleSvg(Builder):
    """A builder to scale svgs.

    Parameters
    ----------
    parameters, args : Tuple[:obj:`pathlib.Path`, str, tuple, list]
        The input parameters (dependencies), including filepaths, for the build
        This tuple should have the 'scale' tuple value specified, which is the
        scale to increase or decrease the image.
        ex: ('scale', 2.0) will double the image size.
    env: :obj:`.builders.Environment`
        The build environment
    """

    action = ("rsvg-convert -f svg -o {builder.outfilepath} "
              "{builder.infilepaths}")
    available = False  # Used as part of sequential builders
    priority = 1000
    required_execs = ('rsvg-convert',)

    infilepath_ext = '.svg'
    outfilepath_ext = '.svg'
    outfilepath_append = '_scale'

    scale = None

    def __init__(self, env, **kwargs):
        super().__init__(env, **kwargs)

        # Set the scale of the svg image
        scale = self.get_parameter('scale')
        try:
            self.scale = float(scale)
        except (ValueError, TypeError):
            msg = ("A scale parameter (float) should be specified for the "
                   "'{}' Builder")
            raise ValueError(msg.format(self.__class__.__name__))

    def run_cmd_args(self):
        args = list(super().run_cmd_args())
        args = [args[0]] + ["-z", str(self.scale)] + args[1:]
        return tuple(args)



def get_version(ver_tuple):
    """Convert a version tuple to a version string that is PEP 440 compliant.

    Parameters
    ----------
    ver_tuple: tuple
        A tuple with five items for the version number.
        - item 1: Major version number
        - item 2: Minor version number
        - item 3: Update number
        - item 4: Release type. Either 'final', 'alpha', 'beta', 'rc
        - item 5: Release number

    Returns
    -------
    str
        The version string.

    Examples
    --------
    >>> get_version((2,1,0, 'final', 3))
    '2.1'
    >>> get_version((2,1,0, 'rc', 3))
    '2.1rc3'
    >>> get_version((2,1,1, 'final', 0))
    '2.1.1'
    >>> get_version((2,1,2, 'alpha', 1))
    '2.1.2a1'
    >>>
    """
    assert len(ver_tuple) == 5
    assert ver_tuple[3] in ('final', 'alpha', 'beta', 'rc')
    ver = '.'.join(map(str, ver_tuple[0:2]))
    if ver_tuple[2] > 0:
        ver += '.{}'.format(ver_tuple[2])
    mapping = {'final': '', 'alpha': 'a', 'beta': 'b', 'rc': 'rc'}
    ver += mapping[ver_tuple[3]]
    if ver_tuple[3] in ('alpha', 'beta', 'rc') and ver_tuple[4] > 0:
        ver += str(ver_tuple[4])
    return ver


# The version tuple.
#
# - Item 0 is the major version, usually reserved for major features and API
#   changes
#
# - Item 1 is the minor revision, usually for standard new features.
#
# - Item 2 is the post release number, usually reserved for bug fixes and
#   updates
#
# - Item 3 is the release type. Either 'alpha', 'beta', 'rc' or 'final'
#
# - Item 4 is the release type number. ex: rc3 is the third release candidate.
#   This number is only non-zero for 'alpha', 'beta' and 'rc' software.
VERSION = (2, 3, 2, 'final', 0)

__version__ = get_version(VERSION)



"""
A manager for labels from headings, figures, equations, tables and other
document components.

Responsibilities
----------------

The labels and label manager are responsible for:

1. *Numbering*. Accurate and consistent formatting of labels and their
   identifer. *e.g.* Fig. 1.
2. *Internal links*. Work in conjunction with the ``@ref`` to correctly link
   to these elements.
3. *Tracking modification times*. Keep track of the modification times for
   the label so that references to those labels can trigger a document
   rendering.
"""
from . import receivers
from .types import ContentLabel, DocumentLabel
from .exceptions import LabelNotFound
from .label_manager import LabelManager

__all__ = ('receivers', 'ContentLabel', 'DocumentLabel', 'LabelNotFound',
           'LabelManager')



"""
A builder to crop pdf files
"""
from .builder import Builder
from .validators import validate_tuple


class PdfCrop(Builder):
    """A builder to crop a pdf and form a cropped pdf.

    Parameters
    ----------
    parameters : Tuple[:obj:`pathlib.Path`, str, tuple, list]
        The input parameters (dependencies), including filepaths, for the build
        This tuple may have the 'crop' tuple value or 'crop_percentage' value
        specified. If specified the borders by the given single percentage
        or the 4 percentages for the left, bottom, right, and top margins.
        ex: ('crop', 10) or ('crop', (10, 20, 10,20) or ('crop_percentage', 10)
    """

    action = "pdf-crop-margins -o {builder.outfilepath} {builder.infilepaths}"
    available = False  # Use as part of sequential builders
    priority = 1000
    required_execs = ('pdf-crop-margins',)

    infilepath_ext = '.pdf'
    outfilepath_ext = 'pdf'
    outfilepath_append = '_crop'

    crop_percentage = None
    offset = None

    def __init__(self, env, **kwargs):

        super().__init__(env, **kwargs)

        # Set the crop for the image
        crop = (self.get_parameter('crop') or
                self.get_parameter('crop_percentage'))
        if crop:
            # Validate the crop arguments, if specified
            crop = (validate_tuple(crop, type=int, length=4,
                                   raise_error=True)
                    if not isinstance(crop, int) else crop)
            self.crop_percentage = crop

        # Set the offset for the image
        offset = self.get_parameter('offset')
        if offset:
            # Validate the crop arguments, if specified
            offset = (validate_tuple(offset, type=int, length=4,
                                     raise_error=True)
                      if not isinstance(offset, int) else offset)
            self.offset = offset

    def run_cmd_args(self):
        args = list(super().run_cmd_args())
        # Add crop percentage options
        if isinstance(self.crop_percentage, int):
            args = [args[0]] + ["-p", str(self.crop_percentage)] + args[1:]
        elif isinstance(self.crop_percentage, tuple):
            args = ([args[0]] +
                    ["-p4", *[str(i) for i in self.crop_percentage]] +
                    args[1:])

        # Add crop offsets
        if isinstance(self.offset, tuple):
            args = ([args[0]] +
                    ["-a4", *[str(i) for i in self.offset]] +
                    args[1:])
        return tuple(args)
